# -*- encoding:utf-8 -*-
"""
    技术线内部执行模块
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import math
import logging
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from scipy import optimize as sco, stats
from sklearn import metrics
from sklearn.cluster import KMeans
from sklearn.linear_model import LinearRegression

# noinspection PyUnresolvedReferences
from ..CoreBu.ABuFixes import xrange
from ..CoreBu.ABuPdHelper import pd_resample
from ..CoreBu import ABuEnv
from ..CoreBu.ABuEnv import EMarketDataSplitMode
from ..UtilBu.ABuProgress import AbuProgress
from ..UtilBu import ABuRegUtil, ABuScalerUtil
from ..MarketBu import ABuSymbolPd

__author__ = '阿布'
__weixin__ = 'abu_quant'

"""模块打印根据环境选择logging.info或者print函数"""
log_func = logging.info if ABuEnv.g_is_ipython else print
"""多项拟合函数寻找阻力，支撑位的poly倍数基数"""
g_upport_resistance_unit = 3


def shift_distance(arr, how, slice_start=0, slice_end=-1, color='r', show=True, show_log=True, ps=True):
    """
    计算序列arr的'位移路程比'，可视化形式为直接三角展示位移，路程直接的关系，返回
    h_distance(三角底边距离), v_distance(三角垂直距离),
    distance(斜边，路程), shift(位移), sd（位移路程比：shift / distance）
    :param arr: numpy array
    :param how: EShiftDistanceHow，计算路程终点点位值使用的计算方法
    :param slice_start: 如果arr是子序列切片，切片start值， int
    :param slice_end: 如果arr是子序列切片，切片end值， int
    :param color: 直角三角边框颜色，str对象 eg：'r', 'red'
    :param show_log: 是否输出位移路程比各个段比值，默认True
    :param show: 是否可视化
    :param ps: 是否立即执行plt.show
    :return: h_distance(三角底边距离), v_distance(三角垂直距离),
            distance(斜边，路程), shift(位移), sd（位移路程比：shift / distance）
    """
    if slice_end == -1:
        # 因为下面还有slice_end - slice_start等操作，所以转换为正切片值
        slice_end = len(arr)
    # 根据slice_start， slice_end做切片
    slice_arr = arr[slice_start:slice_end]

    # 位置0的值即为路程起点点位值
    start_y = slice_arr[0]
    how = shift_distance_how(how)
    # 根据how的方法计算slice_arr出路程终点点位值
    shift_y = how(slice_arr)
    # 直角三角底边distance
    h_distance = slice_end - slice_start
    # 直角三角垂直边distance
    v_distance = abs(shift_y - start_y)
    # 计算斜边值，即路程值
    distance = math.sqrt(h_distance ** 2 + v_distance ** 2)

    # 开始计算位移，首先套上pd.Series做diff结果如下eg形式所示
    diff_slice = pd.Series(slice_arr).diff().fillna(value=0)
    """
        eg: diff_slice
            0     0.0000
            1    -3.7984
            2     7.6824
            3     9.0512
            4     4.3459
            5    -7.7679
                   ...
            44   -3.1482
            45   -1.6426
            46   -2.1216
            47    0.4449
            48   -4.6539
            49    7.1520
    """
    # np.abs(diff_slice).sum计算结果即为位移值
    shift = np.abs(diff_slice).sum()
    # 计算出位移路程比sd
    sd = shift / distance

    if show:
        # 可视化
        start_pos = (slice_start, start_y)
        shift_pos = (slice_end, shift_y)

        # 路程起点点位值 > 路程终点点位值
        if shift_y > start_y:
            end_y = start_y
            top_y = shift_y
            end_pos = (slice_end, end_y)
        else:
            end_y = shift_y
            top_y = start_y
            end_pos = (slice_start, end_y)
        # annotate文字显示两条直接边长度
        plt.annotate('h = {:.2f}, v={:.2f}'.format(h_distance, v_distance), xy=(slice_start, end_y - 8))
        # annotate文字显示路程值
        plt.annotate('distance = {:.2f}'.format(distance), xy=(slice_start, top_y + 15))
        # annotate文字显示位移值
        plt.annotate('shift={:.2f}'.format(shift), xy=(slice_start + 5, top_y + 5))

        # legend显示位移路程比值
        legend = '{}-{} shift/distance: {:.2f}'.format(slice_start, slice_end, sd)

        # 三条边依次相连
        plt.plot([start_pos[0], end_pos[0]],
                 [start_pos[1], end_pos[1]], c=color)

        plt.plot([start_pos[0], shift_pos[0]],
                 [start_pos[1], shift_pos[1]], c=color)

        plt.plot([end_pos[0], shift_pos[0]],
                 [end_pos[1], shift_pos[1]], c=color, label=legend)
        plt.legend(loc=2)
        if ps:
            # 是否立即show
            plt.plot(slice_arr)
            plt.show()
        if show_log:
            log_func(legend)

    # h_distance(三角底边距离), v_distance(三角垂直距离),distance(斜边路程),shift(位移),sd（位移路程比：shift / distance）
    return h_distance, v_distance, distance, shift, sd


def calc_kl_speed(kl, resample=5):
    """
    计算曲线跟随趋势的速度，由于速度是相对的，所以需要在相同周期内与参数曲线进行比对
    :param kl: pd.Series或者numpy序列
    :param resample: 计算速度数值的重采样周期默认5
    :return: 趋势变化敏感速度
    """
    kl_rp = pd_resample(kl, '{}D'.format(resample), how='mean')
    """
        eg: kl_rp
        2011-07-28    1.0027
        2011-08-02    0.9767
        2011-08-07    0.9189
        2011-08-12    0.9174
        2011-08-17    0.9156
    """
    kl_diff = kl_rp.diff()
    """
        eg:
        2011-08-02   -0.0260
        2011-08-07   -0.0578
        2011-08-12   -0.0015
        2011-08-17   -0.0018
    """
    # 二值化 1 -1, 做为速度参数序列慢线
    # noinspection PyTypeChecker
    kl_trend_slow = pd.Series(np.where(kl_diff > 0, 1, -1))
    """
        eg: kl_trend_fast
            0     -1
            1     -1
            2     -1
            3     -1
            4     -1
            5      1
            6      1
            7     -1
            8     -1
            9     -1
    """
    # 慢线向前错一个周期形成快线
    kl_trend_fast = kl_trend_slow.shift(-1)
    """
        egL kl_trend_slow
            0     -1.0
            1     -1.0
            2     -1.0
            3     -1.0
            4      1.0
            5      1.0
            6     -1.0
            7     -1.0
            8     -1.0
            9     -1.0
    """

    kl_trend_ffs = kl_trend_fast[:-1] * kl_trend_slow[:-1]
    """
        慢线 乘 快线 即符号运算
        eg：kl_trend_ffs
        0      1.0
        1      1.0
        2      1.0
        3      1.0
        4     -1.0
        5      1.0
        6     -1.0
        7      1.0
        8      1.0
        9      1.0
    """

    # 符号相同的占所有和的比例即为趋势变化敏感速度值，注意如果没有在相同周期内与参数曲线进行比对的曲线，本速度值即无意义
    speed = kl_trend_ffs.value_counts()[1] / kl_trend_ffs.value_counts().sum()
    return speed


def calc_pair_speed(symbol, benchmark_symbol, resample=5, speed_key='close',
                    start=None, end=None, n_folds=2, show=False):
    """
    参数传递一组symbol对，获取symbol对的金融时间序列数据，根据speed_key获取曲线序列数据，
    分别通过calc_kl_speed计算symbol对的趋势跟随速度，相关性＊敏感度＝敏感度置信度
    :param symbol: eg: 'AU0'
    :param benchmark_symbol: eg: 'XAU'
    :param resample: 计算速度数值的重采样周期默认5
    :param speed_key: 金融时间序列数据取的曲线序列key，默认'close'
    :param start: 获取金融时间序列的start时间
    :param end: 获取金融时间序列的end时间
    :param n_folds: 获取金融时间序列的n_folds参数
    :param show: 是否可视化symbol对的趋势走势对比
    :return: 参数symbol, benchmark_symbol所对应的趋势变化敏感速度数值，以及相关性＊敏感度＝敏感度置信度
    """
    from ..TradeBu import AbuBenchmark
    from ..SimilarBu import ABuCorrcoef, ECoreCorrType

    benchmark = AbuBenchmark(benchmark_symbol, start=start, end=end, n_folds=n_folds, rs=False)
    if benchmark.kl_pd is None:
        return None, None, None
    benchmark_kl = benchmark.kl_pd
    kl = ABuSymbolPd.make_kl_df(symbol, benchmark=benchmark,
                                data_mode=EMarketDataSplitMode.E_DATA_SPLIT_UNDO)
    if kl is None:
        return None, None, None
    # 通过calc_kl_speed计算趋势跟随速度
    kl_speed = calc_kl_speed(kl[speed_key], resample)
    benchmark_kl_speed = calc_kl_speed(benchmark_kl[speed_key], resample)
    # 两个走势的SPERM相关性
    corr = ABuCorrcoef.corr_xy(kl.close, benchmark_kl.close, ECoreCorrType.E_CORE_TYPE_SPERM)

    if show:
        # 可视化symbol对的趋势走势对比
        kl_sl = ABuScalerUtil.scaler_one(kl[speed_key])
        benchmark_kl_sl = ABuScalerUtil.scaler_one(benchmark_kl[speed_key])
        kl_resamp = pd_resample(kl_sl, '{}D'.format(resample), how='mean')
        benchmark_kl_resamp = pd_resample(benchmark_kl_sl, '{}D'.format(resample), how='mean')
        kl_resamp.plot(label='kl', style=['*--'])
        benchmark_kl_resamp.plot(label='benchmark', style=['^--'])
        plt.legend(loc='best')
    # 返回参数symbol, benchmark_symbol所对应的趋势变化敏感速度数值, 以及相关性＊敏感度＝敏感度置信度
    return kl_speed, benchmark_kl_speed, corr


def shift_distance_how(how):
    """
    通过how（EShiftDistanceHow），对应出计算算路程终点点位值使用的计算方法
    注意默认使用shift_distance_close对应标准路程点位值定义方法，其它方法对应的
    路程终点点位值使用的计算方法并非得到最准确的'路程位移比'
    :param how: EShiftDistanceHow对象或者callable即外部可自行设置方法，即计算算路程终点点位值使用的计算方法可自定义
    :return: 计算算路程终点点位值使用的计算方法
    """
    if callable(how):
        # 外部可自行设置方法，即计算算路程终点点位值使用的计算方法可自定义
        return how

    from ..TLineBu.ABuTLine import EShiftDistanceHow

    if how == EShiftDistanceHow.shift_distance_maxmin:
        # 如果p_arr[0] > p_arr[-1]，使用np.min(p_arr)，否则np.max(p_arr)，即上升趋势取max，下跌趋势取min
        how = lambda p_arr: np.min(p_arr) if p_arr[0] > p_arr[-1] else np.max(p_arr)
    elif how == EShiftDistanceHow.shift_distance_close:
        # 对应序列的最后一个点位值，标准路程点位值定义
        how = lambda p_arr: p_arr[-1]
    elif how == EShiftDistanceHow.shift_distance_sum_maxmin:
        # 如果abs(p_arr.max() - p_arr[-1]) > abs(p_arr[-1] - p_arr.min()) 取np.min(p_arr)否则np.max(p_arr)
        # 即最终的点位绝对距离靠近np.min(p_arr)取np.min(p_arr)否则np.max(p_arr)
        how = lambda p_arr: np.min(p_arr) if abs(p_arr.max() - p_arr[-1]) > abs(p_arr[-1] - p_arr.min()) \
            else np.max(p_arr)
    else:
        raise TypeError('how is error how={}'.format(how))
    return how


def regress_trend_channel(arr):
    """
    通过arr计算拟合曲线及上下拟合通道曲线，返回三条拟合曲线，组成拟合通道
    :param arr: numpy array
    :return: y_below, y_fit, y_above
    """
    # 通过ABuRegUtil.regress_y计算拟合曲线和模型reg_mode，不使用缩放参数zoom
    reg_mode, y_fit = ABuRegUtil.regress_y(arr, zoom=False)
    reg_params = reg_mode.params

    x = np.arange(0, len(arr))
    a = reg_params[0]
    b = reg_params[1]
    # 通过argmin寻找出原始序列和拟合序列差值的最小点，差值最小代表点位离拟合曲线远，eg: 100 - 80 < 100 - 90
    min_ind = (arr.T - y_fit).argmin()
    # 根据a, b计算出below值, 注意这里是差，eg: below：100 － 80 ＝ 20
    below = x[min_ind] * b + a - arr[min_ind]
    # 计算x * b + a但- below，即拟合曲线保持相同的斜率整体下移below值
    y_below = x * b + a - below

    # 通过argmax寻找出原始序列和拟合序列差值的最大点，差值最小代表点位离拟合曲线远，eg: 120 - 100 > 120 - 110
    max_ind = (arr.T - y_fit).argmax()
    # 根据a, b计算出above值, 注意这里是差，eg: above 100 - 120 ＝ -20, 即above是负数
    above = x[max_ind] * b + a - arr[max_ind]
    # 计算x * b + a但整天- above，由于above是负数，即相加 即拟合曲线保持相同的斜率整体上移above值
    y_above = x * b + a - above
    return y_below, y_fit, y_above


def bfgs_min_pos(find_min_pos, y_len, linear_interp):
    """
    通过scipy.interpolate.interp1d插值形成的模型，通过sco.fmin_bfgs计算min
    :param find_min_pos: 寻找min的点位值
    :param y_len: 原始序列长度，int
    :param linear_interp: scipy.interpolate.interp1d插值形成的模型
    :return: sco.fmin_bfgs成功找到的值，所有失败的或者异常都返回－1
    """
    try:
        local_min_pos = sco.fmin_bfgs(linear_interp, find_min_pos, disp=False)[0]
    except:
        # 所有失败的或者异常都返回－1
        local_min_pos = -1
    if local_min_pos < 0 or local_min_pos > y_len:
        # 所有失败的或者异常都返回－1
        local_min_pos = -1
    return local_min_pos


def support_resistance_pos(x, support_resistance_y, best_poly=0, label=None):
    """
    分析获取序列阻力位或者支撑位，通过sco.fmin_bfgs寻找阻力位支撑位，阻力位点也是通过sco.fmin_bfgs寻找，
    但是要求传递进来的序列已经是标准化后取反的序列
    eg：
        demean_y = ABuStatsUtil.demean(self.tl)： 首先通过demean将序列去均值
        resistance_y = demean_y * -1 ：阻力位序列要取反
        support_y = demean_y ：支持位序列不需要取反
    sco.fmin_bfgs使用的模型函数为polynomial.Chebyshev多项拟合函数，poly的次数确定
    由ABuRegUtil.search_best_poly得到，即best_poly次多项式拟合回归的趋势曲线可以比较完美的代表原始曲线y的走势，
    为了得到更多的阻力支持位种子点位值，使用：
        np.polynomial.Chebyshev.fit(x, support_resistance_y, best_poly * g_upport_resistance_unit)
        g_upport_resistance_unit默认＝3
    best_poly * 3，即将poly次数又扩大了3倍，可以改变g_upport_resistance_unit获取更多的阻力位支撑位种子点，但
    速度会更慢。
    :param x: 待分析的序列np.array
    :param support_resistance_y:
    :param best_poly: 函数使用者可设置best_poly, 设置后就不使用ABuRegUtil.search_best_poly寻找了
    :param label: 进度条显示的等待文字，str对象
    :return: 阻力位或者支撑位序列
    """
    if best_poly <= 0:
        # 由ABuRegUtil.search_best_poly得到，即best_poly次多项式拟合回归的趋势曲线可以比较完美的代表原始曲线y的走势
        # 注意这里poly_min＝7，即从7次ploy开始，提升效率
        best_poly = ABuRegUtil.search_best_poly(support_resistance_y, poly_min=7, zoom=False, show=False)
        # 根据曲线的长度倍乘best_poly
        best_poly *= int(math.ceil(len(x) / 120))
    best_poly = int(len(x) / 20) if best_poly < int(len(x) / 20) else best_poly

    # 为了得到更多的阻力支持位种子点位值->best_poly * 3，即将poly次数又扩大了3倍
    p = np.polynomial.Chebyshev.fit(x, support_resistance_y, best_poly)

    # 需要使用set，因为需要过滤重复的
    support_resistance = set()
    # 属于耗时操作，构建进度条显示
    with AbuProgress(len(support_resistance_y), 0, label) as progess:
        for index in xrange(0, len(support_resistance_y), 1):
            progess.show(index + 1)
            local_min_pos = int(bfgs_min_pos(index, len(support_resistance_y), p))
            if local_min_pos == -1:
                # 其实主要就是利用这里找不到的情况进行过滤
                continue
            # 将local_min_pos加到集合中
            support_resistance.add(local_min_pos)
    # 为了方便后续api将set转换list
    support_resistance = list(support_resistance)
    return support_resistance


def select_k_support_resistance(support_resistance, thresh=0.06, label='', show=True):
    """
    对阻力位或者支撑位序列从1-序列个数开始聚类，多个聚类器的方差值进行比较，
    通过方差阀值等方法找到最佳聚类个数，最终得到kmean最佳分类器对象
    :param support_resistance: 阻力位或者支撑位序列
    :param thresh: 聚类方差比例阀值，默认0.06
    :param label: 可视化显示的label，主要用来区别阻力位和支撑位的聚类分析结果
    :param show: 是否可视化
    :return: 最佳分类器对象，KMeans类型
    """
    # 阻力位或者支撑位序列从1-序列个数开始聚类
    k_rng = xrange(1, len(support_resistance))
    est_arr = [KMeans(n_clusters=k).fit(support_resistance) for k in k_rng]
    sum_squares = [e.inertia_ for e in est_arr]
    """
        eg: sum_squares 形如
        [568824.24742527946, 63692.462671013389, 23023.512755456246, 11106.460471044047, 4739.3661037023803, 2
        479.206823270833, 1640.5275739375002, 1095.9099614375, 589.28019737500006, 160.28686404166669,
        45.229672916666679, 31.689656250000013, 18.94814375000001, 10.944315625000012, 5.5643156249999999,
        0.6365031249999995]
    """
    # sum_squares[0] = 568824.24742527946 ->
    # 568824.24742527946 / 568824.24742527946, 63692.462671013389 / 568824.24742527946
    diff_squares = [squares / sum_squares[0] for squares in sum_squares]
    diff_squares_pd = pd.Series(diff_squares)

    thresh_pd = diff_squares_pd[diff_squares_pd < thresh]
    if thresh_pd.shape[0] > 0:
        """
            eg：thresh_pd 形如index[0]代表最小的聚类个数值
            2     4.0476e-02
            3     1.9525e-02
            4     8.3319e-03
            5     4.3585e-03
            6     2.8841e-03
            7     1.9266e-03
            8     1.0360e-03
            9     2.8179e-04
        """
        select_k = k_rng[thresh_pd.index[0]]
    else:
        # 没有符合的，就安装最多的个数聚类
        select_k = k_rng[-1]

    # 通过select_k取代分类器
    est = est_arr[select_k - 1]

    if show:
        plt.subplot(211)
        plt.title('{}: elbow method to inform k choice'.format(label))
        # 手肘法可视化最佳聚类个数值，通过silhouette_score
        silhouette_score = [metrics.silhouette_score(support_resistance, e.labels_, metric='euclidean')
                            for e in est_arr[1:]]
        plt.plot(k_rng[1:], silhouette_score, 'b*-')
        plt.grid(True)
        plt.ylabel('{}: Silhouette Coefficient'.format(label))

        # 可视化方差最佳聚类
        plt.subplot(212)
        plt.plot(k_rng, sum_squares, 'b*-')
        plt.grid(True)
        plt.xlabel('{}: k'.format(label))
        plt.ylabel('{}: Sum of Squares'.format(label))
        # 将前面得到的select_k使用圆圈在图示上进行标注
        plt.plot(select_k, sum_squares[select_k - 1], 'ro', markersize=12, markeredgewidth=1.5,
                 markerfacecolor='None', markeredgecolor='r')
        plt.show()
    # 返回最佳分类器对象，KMeans类型
    return est


def support_resistance_predict(x, y, est, support_resistance, is_support, show=True):
    """
    通过最优聚类器est从阻力位或者支撑位序列support_resistance中的序列进行聚类predict
    对聚类结果进行可视化，最终从每个cluster中找到唯一值代表这个分类，如果是支撑位要找到最小值的index
    如果是阻力位找到最大值的index
    :param x: 可迭代序列x
    :param y: 可迭代序列y
    :param est: 最佳分类器对象，KMeans类型
    :param support_resistance: 阻力位或者支撑位序列
    :param is_support: 是否是进行支撑位support_resistance_predict
    :param show: 是否进行可视化
    :return: 返回从每个cluster中找到唯一值代表这个分类的x值组成的序列。list
    """
    support_resistance_k = est.predict(support_resistance)
    """
        eg: support_resistance_k形如：
        array([0, 0, 1, 1, 0, 1, 1, 0, 0, 2, 1, 0, 0, 2, 0, 0, 0], dtype=int32)
    """
    if show:
        # FIXME 这里没有对超出颜色的范围进行处理，如果聚类个数超过颜色，会出错
        colors = np.array(['#FF0054', '#FBD039', '#23C2BC',
                           '#CC99CC', '#CC3399', '#33FF99', '#00CCFF', '#66FF66', '#339999',
                           '#6666CC', '#666666', '#663333', '#660033', '#FF0054', '#FBD039', '#23C2BC',
                           '#CC99CC', '#CC3399', '#33FF99', '#00CCFF', '#66FF66', '#339999',
                           '#6666CC', '#666666', '#663333', '#660033', '#6666CC', '#666666', '#663333',
                           '#660033', '#FF0054', '#FBD039', '#23C2BC', '#CC99CC', '#CC3399', '#33FF99',
                           '#00CCFF', '#66FF66', '#339999'])
        plt.plot(x, y, '-')
        # c=colors[support_resistance_k]即对不同的聚类采用不同的颜色进行标示
        plt.scatter(support_resistance[:, 0], support_resistance[:, 1], c=colors[support_resistance_k], s=60)
        plt.title('{}: k choice'.format('support' if is_support else 'resistance'))
        plt.show()
    d_pd = pd.DataFrame(support_resistance, columns=['x', 'y'])
    d_pd['cluster'] = support_resistance_k
    """
        eg：d_pd形如：
                x        y  cluster
        0     0.0  15.5400        0
        1    97.0  21.3475        0
        2   482.0  22.9625        1
        3   387.0  21.6725        1
        4     5.0  16.2350        0
        5   426.0  23.7650        1
        6   459.0  23.2800        1
        7    48.0  13.1475        0
        8    49.0  13.6700        0
        9   283.0  20.9150        2
        10  497.0  25.2240        1
        11   53.0  13.6537        0
        12   23.0  15.3150        0
        13  280.0  19.9900        2
        14  121.0  18.3475        0
        15   27.0  15.2275        0
        16   95.0  23.9475        0
    """
    k_list = list()

    # 最终从每个cluster中找到唯一值代表这个分类
    for k in set(support_resistance_k):
        if is_support:
            # 如果是支撑位要找到最小值的index
            ind = d_pd[d_pd['cluster'] == k]['y'].argmin()
        else:
            # 如果是阻力位找到最大值的index
            ind = d_pd[d_pd['cluster'] == k]['y'].argmax()
        # 最终的结果只要x的值
        choice_x = int(d_pd.iloc[ind]['x'])
        k_list.append(choice_x)
    # 返回从每个cluster中找到唯一值代表这个分类的x值组成的序列
    return k_list


# noinspection PyArgumentList
def plot_support_resistance_trend(x, y, trend_pos, label, only_last=False, plot_org=False, show=True):
    """
    通过trend_pos绘制阻力线或者支撑线，only_last控制只绘制时间序列中最后一个发现的阻力或支撑，
    plot_org控制是否绘制线段还是直线，plot_org＝True时绘制线段，否则通过LinearRegression进行
    直线绘制
    :param x: 可迭代序列x
    :param y: 可迭代序列y
    :param trend_pos: 趋势点x序列，绘制时需要通过原始序列y获取y[x]值
    :param label: 可视化显示的文字，用来区分阻力位和支撑位
    :param only_last: 控制只绘制时间序列中最后一个发现的阻力或支撑
    :param plot_org: 控制是否绘制线段还是直线，控制是否绘制线段还是直线，
                     plot_org＝True时绘制线段，否则通过LinearRegression进行
    :param show: 是否进行可视化
    """
    # trend_pos类型为list容器
    trend_pos.sort()
    if len(trend_pos) < 2:
        log_func('{} len(trend_pos) < 2 !'.format(label))
        return

    if only_last:
        # 只绘制时间序列中最后一个发现的阻力或支撑，因为前面trend_pos.sort()了，即序列中最后两个元素
        trend_pos = trend_pos[-2:]

    y_trend_arr = []
    for ind, trend_start in enumerate(trend_pos):
        if ind == len(trend_pos) - 1:
            continue
        trend_end = trend_pos[ind + 1]

        x_org = [trend_start, trend_end]
        y_org = [y[trend_start], y[trend_end]]

        # 通过LinearRegression学习线段
        reg = LinearRegression()
        reg.fit(np.array(x_org).reshape(-1, 1), np.array(y_org).reshape(-1, 1))

        x_line = [x[0], x[-1]]
        # predict序列的第一个和最后一个点，即将线段变成了直线，延伸了阻力位，支撑位
        y_line = reg.predict(np.array(x_line).reshape(-1, 1)).reshape(-1, )
        # 把端点结果返回给外面
        y_trend_arr.append(y_line)
        if show:
            if plot_org:
                # plot_org＝True时绘制线段
                plt.plot(x_org, y_org, 'o-', label=label)
            else:
                plt.plot(x_line, y_line, 'o-', label=label)
    if show:
        plt.plot(x, y)
    return y_trend_arr


def skeleton_how(how):
    """
    根据how映射计算数据序列骨架点位的方法
    :param how: ESkeletonHow对象或者callable即外部可自行设置方法，即计算数据序列骨架点位的方法可自定义
    :return:
    """
    if callable(how):
        # callable即外部可自行设置方法
        return how

    from ..TLineBu.ABuTLine import ESkeletonHow
    if how == ESkeletonHow.skeleton_min:
        how_func = np.min
    elif how == ESkeletonHow.skeleton_max:
        how_func = np.max
    elif how == ESkeletonHow.skeleton_mean:
        how_func = np.mean
    elif how == ESkeletonHow.skeleton_median:
        how_func = np.median
    elif how == ESkeletonHow.skeleton_close:
        # 取序列最后一个元素做为采样骨架点位
        how_func = lambda arr: arr[-1]
    elif how == ESkeletonHow.skeleton_triangle:
        # 三角模式骨架点位：确定取最大值，最小值，第三个点位how_func提供
        # 如果np.argmax(arr) > np.argmin(arr)即最大值位置在最小值前面，第三点取序列起点，否则取序列终点
        how_func = lambda arr, start: (start, arr[0]) if np.argmax(arr) > np.argmin(arr) else \
            (len(arr) + start, arr[-1])
    else:
        raise TypeError('how is error how={}'.format(how))
    return how_func


def below_above_gen(x, y):
    """
    (y, x) if x > y else (x, y)
    :param x: 支持比较操作的对象
    :param y: 支持比较操作的对象
    """
    return (y, x) if x > y else (x, y)


def find_percent_point(percents, y):
    """
    可视化技术线比例分割的区域, 针对输入的比例迭代操作后
    分别使用stats.scoreatpercentile和 (y.max() - y.min()) * pt + y.min()两种
    方式进行计算的分割值, 返回对象为比例值为key的字典对象
    eg:
        input:
            percents = (0.1, 0.9)
        output:
            {0.1: (15.732749999999999, 15.5075), 0.9: (31.995000000000005, 34.387500000000003)}

    :param percents: 可迭代序列，eg: (0.1, 0.9), [0.3, 0,4, 0.8]
    :param y: 计算分割线的序列
    :return: 比例值为key的字典对象
    """
    percent_point_dict = {pt: (stats.scoreatpercentile(y, np.round(pt * 100, 1)), (y.max() - y.min()) * pt + y.min())
                          for pt in percents}

    return percent_point_dict


# noinspection PyTypeChecker
def find_golden_point_ex(x, y, show=False):
    """统计黄金分割计算方法，以及对应简单可视化操作"""

    sp382 = stats.scoreatpercentile(y, 38.2)
    sp618 = stats.scoreatpercentile(y, 61.8)
    sp50 = stats.scoreatpercentile(y, 50.0)

    if show:
        # 可视化操作
        plt.plot(x, y)
        plt.axhline(sp50, color='c')
        plt.axhline(sp618, color='r')
        plt.axhline(sp382, color='g')
        _ = plt.setp(plt.gca().get_xticklabels(), rotation=30)
        plt.legend(['TLine', 'sp50', 'sp618', 'sp382'],
                   bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
        plt.show()

    return sp382, sp50, sp618


def find_golden_point(x, y, show=False):
    """视觉黄金分割计算方法，以及对应简单可视化操作"""

    cs_max = y.max()
    cs_min = y.min()

    sp382 = (cs_max - cs_min) * 0.382 + cs_min
    sp618 = (cs_max - cs_min) * 0.618 + cs_min
    sp50 = (cs_max - cs_min) * 0.5 + cs_min
    if show:
        # 可视化操作
        plt.plot(x, y)
        plt.axhline(sp50, color='c')
        plt.axhline(sp618, color='r')
        plt.axhline(sp382, color='g')
        _ = plt.setp(plt.gca().get_xticklabels(), rotation=30)
        plt.legend(['TLine', 'sp50', 'sp618', 'sp382'],
                   bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
        plt.show()

    return sp382, sp50, sp618



# -*- encoding:utf-8 -*-
"""示例ump主裁特征走势拟合角度模块"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from ..TradeBu.ABuMLFeature import AbuFeatureDeg, AbuFeatureDegExtend
from ..MLBu.ABuMLPd import AbuMLPd
from .ABuUmpBase import ump_main_make_xy, BuyUmpMixin
from .ABuUmpMainBase import AbuUmpMainBase

__author__ = '阿布'
__weixin__ = 'abu_quant'


class AbuUmpMainDeg(AbuUmpMainBase, BuyUmpMixin):
    """主裁走势拟合角度特征类，AbuUmpMainBase子类，混入BuyUmpMixin，做为买入ump类"""

    class UmpDegFiter(AbuMLPd):
        """
            内部类，只需要实现make_xy，且使用ump_main_make_xy装饰

            在__init__中通过：
                    self.fiter_cls = self.get_fiter_class()
                    self.fiter = self.fiter_cls(orders_pd=self.orders_pd, **kwarg)
            构造裁判的filter以及重要的self.fiter.df，即pd.DataFrame对象特征
        """

        @ump_main_make_xy
        def make_xy(self, **kwarg):
            """
            make_xy通过装饰器ump_main_make_xy进行二次包装
            这里只需要使用filter选取需要的特征，即从self.order_has_ret中使用filter选取特征列形成df
            :param kwarg: ump_main_make_xy装饰器中使用kwarg
                          kwargs['orders_pd'] 做为必须要有的关键字参数：交易训练集数据，pd.DataFrame对象
                          kwargs['scaler']    做为可选关键字参数：控制在make_xy中返回的特征矩阵数据是否进行标准化处理

            :return: self.order_has_ret中使用filter选取特征列形成deg_df
                     ump_main_make_xy装饰器在make_xy返回deg_df后做转换matrix，形成x，y等工作

                eg: deg_df

                                result  buy_deg_ang42  buy_deg_ang252  buy_deg_ang60  \
                2014-09-24       0          3.378           3.458          3.458
                2014-10-24       0          0.191           2.889          2.809
                2014-10-29       1         -2.026          16.689         -0.761
                2014-10-29       1         -3.427         -11.956         -8.296
                2014-10-29       1         -2.915          39.469         -6.043
                2014-10-29       1         -2.026          16.689         -0.761
                2014-11-03       1          0.103          39.202         -4.614
                2014-11-11       1          8.341          -9.450          0.730
                2014-11-12       0          3.963           6.595         -7.524
                2014-11-26       0         14.052           6.061          7.566
                ...            ...            ...             ...            ...
                2016-03-14       1          4.002         -10.559         -7.992
                2016-03-14       0          0.129          -6.649        -10.880
                2016-03-30       0         13.121          -8.461          4.498
                2016-04-04       1          4.409         -33.097         -6.281
                2016-04-13       0          6.603         -31.459          0.191
                2016-04-14       0          4.611          18.428          3.134
                2016-04-15       0          4.238         -13.247          4.693
                2016-04-15       0          4.238         -13.247          4.693
                2016-04-29       1          1.445          16.266          4.615
                2016-04-29       1          1.445          16.266          4.615

                            buy_deg_ang21
                2014-09-24          1.818
                2014-10-24         -1.089
                2014-10-29          1.980
                2014-10-29          6.507
                2014-10-29          7.046
                2014-10-29          1.980
                2014-11-03         10.125
                2014-11-11         12.397
                2014-11-12          6.671
                2014-11-26         12.494
                ...                   ...
                2016-03-14          9.324
                2016-03-14          5.201
                2016-03-30          4.070
                2016-04-04          5.618
                2016-04-13          4.457
                2016-04-14          0.733
                2016-04-15          1.162
                2016-04-15          1.162
                2016-04-29         -1.115
                2016-04-29         -1.115
            """
            # regex='result|buy_deg_ang42|buy_deg_ang252|buy_deg_ang60|buy_deg_ang21'
            regex = 'result|{}'.format(
                '|'.join(AbuFeatureDeg().get_feature_ump_keys(ump_cls=AbuUmpMainDeg)))
            # noinspection PyUnresolvedReferences
            deg_df = self.order_has_ret.filter(regex=regex)
            return deg_df

    def get_predict_col(self):
        """
        主裁走势拟合角度特征keys：['buy_deg_ang42', 'buy_deg_ang252', 'buy_deg_ang60', 'buy_deg_ang21']
        :return: ['buy_deg_ang42', 'buy_deg_ang252', 'buy_deg_ang60', 'buy_deg_ang21']
        """
        return AbuFeatureDeg().get_feature_ump_keys(ump_cls=AbuUmpMainDeg)

    def get_fiter_class(self):
        """
        主裁特征走势拟合角度返回的AbuMLPd子类：AbuUmpMainDeg.UmpDegFiter
        :return: AbuUmpMainDeg.UmpDegFiter
        """
        return AbuUmpMainDeg.UmpDegFiter

    @classmethod
    def class_unique_id(cls):
        """
        具体ump类关键字唯一名称，类方法：return 'deg_main'
        主要针对外部user设置自定义ump使用, 需要user自己保证class_unique_id的唯一性，内部不做检测
        具体使用见ABuUmpManager中extend_ump_block方法
        """
        return 'deg_main'


class AbuUmpMainDegExtend(AbuUmpMainBase, BuyUmpMixin):
    """主裁使用新的视角来决策交易，AbuUmpMainBase子类，混入BuyUmpMixin，做为买入ump类"""

    class UmpExtendFeatureFiter(AbuMLPd):
        @ump_main_make_xy
        def make_xy(self, **kwarg):
            # 这里使用get_feature_ump_keys，只需要传递当前类名称即可，其根据是买入ump还是卖出ump返回对应特征列
            col = AbuFeatureDegExtend().get_feature_ump_keys(ump_cls=AbuUmpMainDegExtend)
            regex = 'result|{}'.format('|'.join(col))
            extend_deg_df = self.order_has_ret.filter(regex=regex)
            return extend_deg_df

    def get_predict_col(self):
        # 这里使用get_feature_ump_keys，只需要传递当前类名称即可，其根据是买入ump还是卖出ump返回对应特征列
        col = AbuFeatureDegExtend().get_feature_ump_keys(ump_cls=AbuUmpMainDegExtend)
        return col

    def get_fiter_class(self):
        return AbuUmpMainDegExtend.UmpExtendFeatureFiter

    @classmethod
    def class_unique_id(cls):
        return 'extend_main_deg'



# -*- encoding:utf-8 -*-

"""
股价的标准差及其信赖区间，从而确定股价的波动范围及未来走势，利用波带显示股价的安全高低价位，因而也被称为布林带。
其上下限范围不固定，随股价的滚动而变化。布林指标和麦克指标MIKE一样同属路径指标，股价波动在上限和下限的区间之内，
这条带状区的宽窄，随着股价波动幅度的大小而变化，股价涨跌幅度加大时，带状区变宽，涨跌幅度狭小盘整时，带状区则变窄

计算公式
中轨线=N日的移动平均线
上轨线=中轨线+nb_dev * N日的移动标准差
下轨线=中轨线－nb_dev * N日的移动标准差
（nb_dev为参数，可根据股票的特性来做相应的调整，一般默为2）
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from .ABuNDBase import plot_from_order, g_calc_type, ECalcType
from ..CoreBu.ABuPdHelper import pd_rolling_mean, pd_rolling_std

__author__ = '阿布'
__weixin__ = 'abu_quant'


# noinspection PyUnresolvedReferences
def _calc_boll_from_ta(prices, time_period=20, nb_dev=2):
    """
    使用talib计算boll, 即透传talib.BBANDS计算结果
    :param prices: 收盘价格序列，pd.Series或者np.array
    :param time_period: boll的N值默认值20，int
    :param nb_dev: boll的nb_dev值默认值2，int
    :return: tuple(upper, middle, lower)
    """
    import talib
    if isinstance(prices, pd.Series):
        prices = prices.values

    upper, middle, lower = talib.BBANDS(
        prices,
        timeperiod=time_period,
        nbdevup=nb_dev,
        nbdevdn=nb_dev,
        matype=0)

    return upper, middle, lower


def _calc_boll_from_pd(prices, time_period=20, nb_dev=2):
    """
    通过boll公式手动计算boll
    :param prices: 收盘价格序列，pd.Series或者np.array
    :param time_period: boll的N值默认值20，int
    :param nb_dev: boll的nb_dev值默认值2，int
    :return: tuple(upper, middle, lower)
    """
    # 中轨线 = N日的移动平均线
    middle = pd_rolling_mean(prices, window=time_period, min_periods=time_period)
    # N日的移动标准差
    n_std = pd_rolling_std(prices, window=20, center=False)
    # 上轨线=中轨线+nb_dev * N日的移动标准差
    upper = middle + nb_dev * n_std
    # 下轨线 = 中轨线－nb_dev * N日的移动标准差
    lower = middle - nb_dev * n_std

    # noinspection PyUnresolvedReferences
    return upper.values, middle.values, lower.values


"""通过在ABuNDBase中尝试import talib来统一确定指标计算方式, 外部计算只应该使用calc_boll"""
calc_boll = _calc_boll_from_pd if g_calc_type == ECalcType.E_FROM_PD else _calc_boll_from_ta


def plot_boll_from_klpd(kl_pd, with_points=None, with_points_ext=None, **kwargs):
    """
    封装plot_boll，绘制收盘价格，boll（upper, middle, lower）曲线
    :param kl_pd: 金融时间序列，pd.DataFrame对象
    :param with_points: 这里的常规用途是传入买入order, with_points=buy_index=pd.to_datetime(orders['buy_date']))
    :param with_points_ext: 这里的常规用途是传入卖出order, with_points_ext=sell_index=pd.to_datetime(orders['sell_date']))
    :param kwargs: 绘制技术指标需要的其它关键字参数，ime_period, nb_dev, 最终透传给plot_boll
    """
    plot_boll(kl_pd.close, kl_pd.index, with_points=with_points, with_points_ext=with_points_ext, **kwargs)


def plot_boll_from_order(order, date_ext=120, **kwargs):
    """
    封装ABuNDBase中的plot_from_order与模块中绘制技术指标的函数，完成技术指标可视化及标注买入卖出点位
    :param order: AbuOrder对象转换的pd.DataFrame对象or pd.Series对象
    :param date_ext: int对象 eg. 如交易在2015-06-01执行，如date_ext＝120，择start向前推120天，end向后推120天
    :param kwargs: 绘制技术指标需要的其它关键字参数，ime_period, nb_dev, 最终透传给plot_boll
    """
    return plot_from_order(plot_boll_from_klpd, order, date_ext, **kwargs)


def plot_boll(prices, kl_index, with_points=None, with_points_ext=None, time_period=20, nb_dev=2):
    """
    绘制收盘价格，以及对应的boll曲线，如果有with_points点位标注，使用竖线标注

    :param prices: 收盘价格序列，pd.Series或者np.array
    :param kl_index: pd.Index时间序列
    :param with_points: 这里的常规用途是传入买入order, with_points=buy_index=pd.to_datetime(orders['buy_date']))
    :param with_points_ext: 这里的常规用途是传入卖出order, with_points_ext=sell_index=pd.to_datetime(orders['sell_date']))
    :param time_period: 计算boll使用的n日参数，默认20
    :param nb_dev: 计算boll使用的nb_dev参数，默认2
    :return:
    """
    upper, middle, lower = np.array(calc_boll(prices, time_period, nb_dev))

    plt.figure(figsize=[14, 7])

    plt.plot(kl_index, prices, label='close price')
    plt.plot(kl_index, upper, label='upper')
    plt.plot(kl_index, middle, label='middle')
    plt.plot(kl_index, lower, label='lower')

    # with_points和with_points_ext的点位使用竖线标注
    if with_points is not None:
        plt.axvline(with_points, color='green', linestyle='--')

    if with_points_ext is not None:
        plt.axvline(with_points_ext, color='red')

    plt.grid(True)
    plt.legend(loc='best')
    plt.show()



# coding=utf-8
"""
    数据对外接口模块，其它模块需要数据都只应该使用ABuSymbolPd, 不应涉及其它内部模块的使用
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import logging
from collections import Iterable

import pandas as pd

from .ABuDataSource import kline_pd
from ..MarketBu.ABuDataCache import save_kline_df
from .ABuSymbol import Symbol
from .ABuMarket import split_k_market
from ..CoreBu import ABuEnv
from ..CoreBu.ABuEnv import EMarketDataFetchMode, EDataCacheType
from ..CoreBu.ABuFixes import partial, ThreadPoolExecutor
# noinspection PyUnresolvedReferences
from ..CoreBu.ABuFixes import range
from ..CoreBu.ABuDeprecated import AbuDeprecated
from ..IndicatorBu import ABuNDAtr as Atr
from ..UtilBu import ABuDateUtil
from ..UtilBu.ABuFileUtil import batch_h5s
from ..UtilBu.ABuProgress import AbuMulPidProgress, do_clear_output
from ..CoreBu.ABuParallel import delayed, Parallel
from ..CoreBu.ABuFixes import six
# from ..UtilBu.ABuThreadPool import AbuThreadPoolExecutor

__author__ = '阿布'
__weixin__ = 'abu_quant'


def _benchmark(df, benchmark, symbol):
    """
    在内部使用kline_pd获取金融时间序列pd.DataFrame后，如果参数中
    基准benchmark（pd.DataFrame对象）存在，使用基准benchmark的
    时间范围切割kline_pd返回的金融时间序列
    :param df: 金融时间序列pd.DataFrame对象
    :param benchmark: 资金回测时间标尺，AbuBenchmark实例对象
    :param symbol: Symbol对象
    :return: 使用基准的时间范围切割返回的金融时间序列
    """
    if len(df.index & benchmark.kl_pd.index) <= 0:
        # 如果基准benchmark时间范围和输入的df没有交集，直接返回None
        return None

    # 两个金融时间序列通过loc寻找交集
    kl_pd = df.loc[benchmark.kl_pd.index]
    # nan的date个数即为不相交的个数
    nan_cnt = kl_pd['date'].isnull().value_counts()
    # 两个金融序列是否相同的结束日期
    same_end = df.index[-1] == benchmark.kl_pd.index[-1]
    # 两个金融序列是否相同的开始日期
    same_head = df.index[0] == benchmark.kl_pd.index[0]

    # 如果nan_cnt即不相交个数大于benchmark基准个数的1/3，放弃
    base_keep_div = 3
    if same_end or same_head:
        # 如果两个序列有相同的开始或者结束改为1/2，也就是如果数据头尾日起的标尺有一个对的上的话，放宽na数量
        base_keep_div = 2
    if same_end and same_head:
        # 如果两个序列同时有相同的开始和结束改为1，也就是如果数据头尾日起的标尺都对的上的话，na数量忽略不计
        base_keep_div = 1

    if symbol.is_a_stock():
        # 如果是A股市场的目标，由于停盘频率和周期都会长与其它市场所以再放宽一些
        base_keep_div *= 0.7

    if nan_cnt.index.shape[0] > 0 and nan_cnt.index.tolist().count(True) > 0 \
            and nan_cnt.loc[True] > benchmark.kl_pd.shape[0] / base_keep_div:
        # nan 个数 > 基准base_keep_div分之一放弃
        return None

    # 来到这里说明没有放弃，那么就填充nan
    # 首先nan的交易量是0
    kl_pd.volume.fillna(value=0, inplace=True)
    # nan的p_change是0
    kl_pd.p_change.fillna(value=0, inplace=True)
    # 先把close填充了，然后用close填充其它的
    kl_pd.close.fillna(method='pad', inplace=True)
    kl_pd.close.fillna(method='bfill', inplace=True)
    # 用close填充open
    kl_pd.open.fillna(value=kl_pd.close, inplace=True)
    # 用close填充high
    kl_pd.high.fillna(value=kl_pd.close, inplace=True)
    # 用close填充low
    kl_pd.low.fillna(value=kl_pd.close, inplace=True)
    # 用close填充pre_close
    kl_pd.pre_close.fillna(value=kl_pd.close, inplace=True)

    # 细节nan处理完成后，把剩下的nan都填充了
    kl_pd = kl_pd.fillna(method='pad')
    # bfill再来一遍只是为了填充最前面的nan
    kl_pd.fillna(method='bfill', inplace=True)

    # pad了数据所以，交易日期date的值需要根据time index重新来一遍
    kl_pd['date'] = [int(ts.date().strftime("%Y%m%d")) for ts in kl_pd.index]
    kl_pd['date_week'] = kl_pd['date'].apply(lambda x: ABuDateUtil.week_of_date(str(x), '%Y%m%d'))

    return kl_pd


def _make_kl_df(symbol, data_mode, n_folds, start, end, benchmark, save):
    """
    针对一个symbol进行数据获取，内部使用kline_pd从本地加载或者指定数据源进行网络请求
    :param symbol: str对象 or Symbol对象
    :param data_mode: EMarketDataSplitMode enum对象
    :param n_folds: 请求几年的历史回测数据int
    :param start: 请求的开始日期 str对象
    :param end: 请求的结束日期 str对象
    :param benchmark: 资金回测时间标尺，AbuBenchmark实例对象
    :param save: 是否进行网络获取数据后，直接进行本地保存
    :return: (df: 金融时间序列pd.DataFrame对象，save_kl_key: 提供外部进行保存)
    """
    df, save_kl_key = kline_pd(symbol, data_mode, n_folds=n_folds, start=start, end=end, save=save)
    if df is not None and df.shape[0] == 0:
        # 把行数＝0的归结为＝None, 方便后续统一处理
        df = None

    if benchmark is not None and df is not None:
        # 如果有标尺，进行标尺切割，进行标尺切割后也可能变成none
        temp_symbol = save_kl_key[0]
        df = _benchmark(df, benchmark, temp_symbol)

    if df is not None:
        # 规避重复交易日数据风险，subset只设置date做为滤除重复
        df.drop_duplicates(subset=['date'], inplace=True)
        # noinspection PyProtectedMember
        if not ABuEnv._g_enable_example_env_ipython or 'atr14' not in df.columns or 'atr21' not in df.columns:
            # 非沙盒环境计算, 或者是沙盒但数据本身没有atr14，atr21
            calc_atr(df)
        # 根据df长度重新进行key计算
        df['key'] = list(range(0, len(df)))
        temp_symbol = save_kl_key[0]
        df.name = temp_symbol.value
    return df, save_kl_key


def _kl_df_dict_parallel(choice_symbols, data_mode, n_folds, start, end, benchmark):
    """
    多进程或者多线程被委托的任务函数，多任务批量获取时间序列数据
    :param choice_symbols: symbol序列
    :param data_mode: EMarketDataSplitMode enum对象
    :param n_folds: 请求几年的历史回测数据int
    :param start: 请求的开始日期 str对象
    :param end: 请求的结束日期 str对象
    :param benchmark: 资金回测时间标尺，AbuBenchmark实例对象
    :return: df_dict字典中key=请求symbol的str对象，value＝(save_kl_key: 提供外部进行保存, df: 金融时间序列pd.DataFrame对象)
    """
    df_dict = {}
    # 注意save=False

    # 启动多进程进度条，如果是多线程暂时也启动了AbuMulPidProgress，需优化
    with AbuMulPidProgress(len(choice_symbols), 'kl_df parallel complete') as progress:
        for epoch, symbol in enumerate(choice_symbols):
            # 迭代choice_symbols进行_make_kl_df, 注意_make_kl_df的参数save=False，即并行获取，不在内部save，要在外部save
            df, key_tuple = _make_kl_df(symbol, data_mode=data_mode,
                                        n_folds=n_folds, start=start, end=end, benchmark=benchmark, save=False)
            if isinstance(key_tuple, tuple) and len(key_tuple) == 3:
                # key=请求symbol的str对象，value＝(save_kl_key: 提供外部进行保存, df: 金融时间序列pd.DataFrame对象)
                df_dict[key_tuple[0].value] = (key_tuple, df)
            progress.show(epoch + 1)

    return df_dict


def kl_df_dict_parallel(symbols, data_mode=ABuEnv.EMarketDataSplitMode.E_DATA_SPLIT_SE,
                        n_folds=2, start=None, end=None, benchmark=None, n_jobs=16, save=True, how='thread'):
    """
    多进程或者多线程对外执行函数，多任务批量获取时间序列数据
    :param symbols: symbol序列
    :param data_mode: EMarketDataSplitMode enum对象
    :param n_folds: 请求几年的历史回测数据int
    :param start: 请求的开始日期 str对象
    :param end: 请求的结束日期 str对象
    :param benchmark: 资金回测时间标尺，AbuBenchmark实例对象
    :param n_jobs: 并行的任务数，对于进程代表进程数，线程代表线程数
    :param save: 是否统一进行批量保存，即在批量获取金融时间序列后，统一进行批量保存，默认True
    :param how: process：多进程，thread：多线程，main：单进程单线程
    """

    # TODO Iterable和six.string_types的判断抽出来放在一个模块，做为Iterable的判断来使用
    if not isinstance(symbols, Iterable) or isinstance(symbols, six.string_types):
        # symbols必须是可迭代的序列对象
        raise TypeError('symbols must a Iterable obj!')
    # 可迭代的symbols序列分成n_jobs个子序列
    parallel_symbols = split_k_market(n_jobs, market_symbols=symbols)
    # 使用partial对并行函数_kl_df_dict_parallel进行委托
    parallel_func = partial(_kl_df_dict_parallel, data_mode=data_mode, n_folds=n_folds, start=start, end=end,
                            benchmark=benchmark)
    # 因为切割会有余数，所以将原始设置的进程数切换为分割好的个数, 即32 -> 33 16 -> 17
    n_jobs = len(parallel_symbols)
    if how == 'process':
        """
            mac os 10.9 以后的并行加上numpy不是crash就是进程卡死，不要用，用thread
        """
        if ABuEnv.g_is_mac_os:
            logging.info('mac os 10.9 parallel with numpy crash or dead!!')

        parallel = Parallel(
            n_jobs=n_jobs, verbose=0, pre_dispatch='2*n_jobs')
        df_dicts = parallel(delayed(parallel_func)(choice_symbols)
                            for choice_symbols in parallel_symbols)
    elif how == 'thread':
        # 通过ThreadPoolExecutor进行线程并行任务
        with ThreadPoolExecutor(max_workers=n_jobs) as pool:
            k_use_map = True
            if k_use_map:
                df_dicts = list(pool.map(parallel_func, parallel_symbols))
            else:
                futures = [pool.submit(parallel_func, symbols) for symbols in parallel_symbols]
                df_dicts = [future.result() for future in futures if future.exception() is None]
    elif how == 'main':
        # 单进程单线程
        df_dicts = [parallel_func(symbols) for symbols in parallel_symbols]
    else:
        raise TypeError('ONLY process OR thread!')

    if save:
        # 统一进行批量保存
        h5s_fn = ABuEnv.g_project_kl_df_data if ABuEnv.g_data_cache_type == EDataCacheType.E_DATA_CACHE_HDF5 else None

        @batch_h5s(h5s_fn)
        def _batch_save():
            for df_dict in df_dicts:
                # 每一个df_dict是一个并行的序列返回的数据
                for ind, (key_tuple, df) in enumerate(df_dict.values()):
                    # (key_tuple, df)是保存kl需要的数据, 迭代后直接使用save_kline_df
                    save_kline_df(df, *key_tuple)
                    if df is not None:
                        print("save kl {}_{}_{} {}/{}".format(key_tuple[0].value, key_tuple[1], key_tuple[2], ind,
                                                              df.shape[0]))
                # 完成一层循环一次，即批量保存完一个并行的序列返回的数据后，进行清屏
                do_clear_output()

        _batch_save()
    return df_dicts


# noinspection PyDeprecation
def make_kl_df(symbol, data_mode=ABuEnv.EMarketDataSplitMode.E_DATA_SPLIT_SE,
               n_folds=2, start=None, end=None, benchmark=None, show_progress=True, parallel=False, parallel_save=True):
    """
    外部获取金融时间序列接口
    eg: n_fold=2, start=None, end=None ，从今天起往前数两年
        n_fold=2, start='2015-02-14', end=None， 从2015-02-14到现在，n_fold无效
        n_fold=2, start=None, end='2016-02-14'，从2016-02-14起往前数两年
        n_fold=2, start='2015-02-14', end='2016-02-14'，从start到end

    :param data_mode: EMarketDataSplitMode对象
    :param symbol: list or Series or str or Symbol
                    e.g :['TSLA','SFUN'] or 'TSLA' or Symbol(MType.US,'TSLA')
    :param n_folds: 请求几年的历史回测数据int
    :param start: 请求的开始日期 str对象
    :param end: 请求的结束日期 str对象
    :param benchmark: 资金回测时间标尺，AbuBenchmark实例对象
    :param show_progress: 是否显示进度条
    :param parallel: 是否并行获取
    :param parallel_save: 是否并行后进行统一批量保存
    """

    if isinstance(symbol, (list, tuple, pd.Series, pd.Index)):
        # 如果symbol是可迭代的序列对象，最终返回三维面板数据pd.Panel
        panel = dict()
        if parallel:
            # 如果并行获取
            if ABuEnv.g_data_fetch_mode != EMarketDataFetchMode.E_DATA_FETCH_FORCE_NET \
                    and ABuEnv.g_data_cache_type == EDataCacheType.E_DATA_CACHE_HDF5:
                # 只能针对非hdf5存贮形式下或者针对force_net，因为hdf5多线程读容易卡死
                raise RuntimeError('make_kl_df just suit force net or not hdf5 store!')

            df_dicts = kl_df_dict_parallel(symbol, data_mode=data_mode, n_folds=n_folds, start=start, end=end,
                                           benchmark=benchmark, save=parallel_save, how='thread')
            for df_dict in df_dicts:
                for key_tuple, df in df_dict.values():
                    if df is None or df.shape[0] == 0:
                        continue
                    # 即丢弃原始df_dict保存金融时间序列时使用的save_kl_key，只保留df，赋予panel
                    panel[key_tuple[0].value] = df
        else:
            def _batch_make_kl_df():
                with AbuMulPidProgress(len(symbol), '_make_kl_df complete') as progress:
                    for pos, _symbol in enumerate(symbol):
                        _df, _ = _make_kl_df(_symbol, data_mode=data_mode,
                                             n_folds=n_folds, start=start, end=end, benchmark=benchmark, save=True)
                        if show_progress:
                            progress.show()
                        # TODO 做pd.Panel数据应该保证每一个元素的行数和列数都相等，不是简单的有数据就行
                        if _df is None or _df.shape[0] == 0:
                            continue

                        panel[symbol[pos]] = _df

            _batch_make_kl_df()
        # TODO pd.Panel过时
        return pd.Panel(panel)

    elif isinstance(symbol, Symbol) or isinstance(symbol, six.string_types):
        # 对单个symbol进行数据获取
        df, _ = _make_kl_df(symbol, data_mode=data_mode,
                            n_folds=n_folds, start=start, end=end, benchmark=benchmark, save=True)
        return df
    else:
        raise TypeError('symbol type is error')


def get_price(symbol, start_date=None, end_date=None):
    """
    通过make_kl_df获取金融时间序列后，只保留收盘价格，只是为了配合主流回测平台接口名称，适配使用
    :param symbol: str对象或Symbol对象
    :param start_date: 请求的开始日期 str对象
    :param end_date: 请求的结束日期 str对象
    :return: 金融时间序列pd.DataFrame对象只有一个price列
    """
    df = make_kl_df(symbol, start=start_date, end=end_date)
    if df is not None:
        df = df.filter(['close'])
        # 为了配合主流回测平台适配
        return df.rename(columns={'close': 'price'})


def combine_pre_kl_pd(kl_pd, n_folds=1):
    """
    通过传人一个kl_pd获取这个kl_pd之前n_folds年时间的kl，默认n_folds=1,
    eg. kl_pd 从2014-07-26至2016-07-26，首先get 2013-07-26至2014-07-25
    之后合并两段数据，最终返回的数据为2013-07-26至2016-07-26
    :param kl_pd: 金融时间序列pd.DataFrame对象
    :param n_folds: 获取之前n_folds年的数据
    :return: 结果是和输入kl_pd合并后的总kl
    """

    # 获取kl_pd的起始时间
    end = ABuDateUtil.timestamp_to_str(kl_pd.index[0])
    # kl_pd的起始时间做为end参数通过make_kl_df和n_folds参数获取之前的一段时间序列
    pre_kl_pd = make_kl_df(kl_pd.name, data_mode=ABuEnv.EMarketDataSplitMode.E_DATA_SPLIT_SE, n_folds=n_folds,
                           end=end)
    # 再合并两段时间序列，pre_kl_pd[:-1]跳过重复的end
    combine_kl = kl_pd if pre_kl_pd is None else pre_kl_pd[:-1].append(kl_pd)
    # 根据combine_kl长度重新进行key计算
    combine_kl['key'] = list(range(0, len(combine_kl)))
    return combine_kl


def calc_atr(kline_df):
    """
    为输入的kline_df金融时间序列计算atr21和atr14，计算结果直接加到kline_df的atr21列和atr14列中
    :param kline_df: 金融时间序列pd.DataFrame对象
    """
    kline_df['atr21'] = 0
    if kline_df.shape[0] > 21:
        # 大于21d计算atr21
        kline_df['atr21'] = Atr.atr21(kline_df['high'].values, kline_df['low'].values, kline_df['pre_close'].values)
        # 将前面的bfill
        kline_df['atr21'].fillna(method='bfill', inplace=True)
    kline_df['atr14'] = 0
    if kline_df.shape[0] > 14:
        # 大于14d计算atr14
        kline_df['atr14'] = Atr.atr14(kline_df['high'].values, kline_df['low'].values, kline_df['pre_close'].values)
        # 将前面的bfill
        kline_df['atr14'].fillna(method='bfill', inplace=True)


@AbuDeprecated('only for old abu!')
def get_n_year(kl_pd, from_year, get_year=1, how='ff'):
    """
    获取pd中第n年切片数据, Deprecated
    :param kl_pd: 金融时间序列pd.DataFrame对象
    :param from_year: form 1开始纠错0 to 1
    :param get_year: 要几年的数据1就是1年，0.5半年 默认1 year支持0.1 to inf
    :param how:='bf' 从后向前切 ='ff' 从前向后切
    """
    td = ABuEnv.g_market_trade_year
    n_n_year = int(kl_pd.shape[0] / td)
    if from_year == 0:
        from_year = 1
        logging.info('get_n_year form num 1 so you pass 0 covert to 1')
    if from_year > n_n_year:
        raise ValueError('get_n_year n_year > n_n_year!')

    get_days = int(get_year * td)

    if how == 'ff':
        # 从前向后切
        st = int((from_year - 1) * td)
        ed = st + get_days if st + get_days < kl_pd.shape[0] else kl_pd.shape[0]
    elif how == 'bf':
        ed = kl_pd.shape[0] - int((from_year - 1) * td)
        st = ed - get_days - 1 if ed - get_days > 0 else 0
    else:
        raise TypeError('error direction input!')

    ys = slice(st, ed)
    ret_pd = kl_pd[ys]
    if hasattr(kl_pd, 'name'):
        ret_pd.name = kl_pd.name
    return ret_pd



# coding=utf-8
"""
    数据源基础模块
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from abc import ABCMeta, abstractmethod

from ..MarketBu.ABuSymbol import Symbol
from ..CoreBu.ABuEnv import EMarketTargetType
from ..CoreBu.ABuFixes import six
from ..UtilBu import ABuDateUtil


# noinspection PyUnresolvedReferences
class SupportMixin(object):
    """混入类，声明数据源支持的市场，以及检测是否支持市场"""

    def _support_market(self):
        """声明数据源支持的市场，默认声明支持美股，港股，a股"""
        return [EMarketTargetType.E_MARKET_TARGET_US, EMarketTargetType.E_MARKET_TARGET_HK,
                EMarketTargetType.E_MARKET_TARGET_CN]

    def check_support(self, symbol=None, rs=True):
        """
        检测参数symbol对象或者内部self._symbol是否被数据源支持
        :param symbol: 外部可设置检测symbol对象，Symbol对象，EMarketTargetType对象或字符串对象
        :param rs: 如果数据源不支持，是否抛出异常，默认抛出
        :return: 返回是否支持 bool
        """
        if symbol is None:
            symbol = self._symbol

        if isinstance(symbol, six.string_types):
            # 如果是str，使用_support_market返回的value组成字符串数组，进行成员测试
            if symbol in [market.value for market in self._support_market()]:
                return True
        else:
            if isinstance(symbol, Symbol):
                # Symbol对象取market
                market = symbol.market
            elif isinstance(symbol, EMarketTargetType):
                market = symbol
            else:
                raise TypeError('symbol type is Symbol or str!!')
            # _support_market序列进行成员测试
            if market in self._support_market():
                return True

        if rs:
            #  根据rs设置，如果数据源不支持，抛出异常
            raise TypeError('{} don\'t support {}!'.format(self.__class__.__name__, symbol))
        return False


class BaseMarket(object):
    """数据源基础市场基类"""

    # 预先设置模拟手机请求的device
    K_DEV_MODE_LIST = ["A0001", "OPPOR9", "OPPOR9", "VIVOX5",
                       "VIVOX6", "VIVOX6PLUS", "VIVOX9", "VIVOX9PLUS"]
    # 预先设置模拟手机请求的os version
    K_OS_VERSION_LIST = ["4.3", "4.2.2", "4.4.2", "5.1.1"]
    # 预先设置模拟手机请求的屏幕大小
    K_PHONE_SCREEN = [[1080, 1920]]

    def __init__(self, symbol):
        """
        :param symbol: Symbol类型对象
        """
        if not isinstance(symbol, Symbol):
            raise TypeError('symbol is not type Symbol')

        self._symbol = symbol

    # noinspection PyMethodMayBeStatic
    def req_time(self):
        """请求时间seconds模拟"""
        tm = int(ABuDateUtil.time_seconds() * 1000)
        return tm

    @classmethod
    def _fix_kline_pd_se(cls, kl_df, n_folds, start=None, end=None):
        """
        删除多余请求数据，即重新根据start，end或n_folds参数进行金融时间序列切割
        :param kl_df: 金融时间序列切割pd.DataFrame对象
        :param n_folds: n_folds年的数据
        :return: 删除多余数据，规则后的pd.DataFrame对象
        """
        if kl_df is None:
            return kl_df

        # 从start和end中切片
        if start is not None:
            # 有start转换为int，使用kl_df中的date列进行筛选切割
            start = ABuDateUtil.date_str_to_int(start)
            kl_df = kl_df[kl_df.date >= start]
            if end is not None:
                # 有end转换为int，使用kl_df中的date列进行筛选切割
                end = ABuDateUtil.date_str_to_int(end)
                kl_df = kl_df[kl_df.date <= end]
        else:
            # 根据n_folds构造切片的start
            start = ABuDateUtil.begin_date(365 * n_folds)
            start = ABuDateUtil.date_str_to_int(start)
            # 使用kl_df中的date列进行筛选切割
            kl_df = kl_df[kl_df.date >= start]
        return kl_df

    @classmethod
    def _fix_kline_pd_zero(cls, kl_df):
        """
        修复金融时间序列切中的异常点，比如价格为0的点，注意只能用在确定会有异常点发生的市场，
        比如期权期货市场进入交割日，否则不能随意进行修复
        :param kl_df: 金融时间序列切割pd.DataFrame对象
        :return: 修复后的金融时间序列pd.DataFrame对象
        """

        if kl_df is None:
            return kl_df

        def fix_zero(trade_day):
            """对高开低收为0的价格认为是异常，使用昨天的价格替换"""
            if trade_day.close == 0:
                # 收盘异常，如果今天的low还是>0的使用low，否则使用昨天收盘价格
                trade_day.close = trade_day.low if trade_day.low > 0 else trade_day.pre_close

            # 由于已fix了trade_day.close，所以high，low，open使用trade_day.close fix
            if trade_day.high == 0:
                trade_day.high = trade_day.close
            if trade_day.low == 0:
                trade_day.low = trade_day.close
            if trade_day.open == 0:
                trade_day.open = trade_day.close
            return trade_day

        kl_df = kl_df.apply(fix_zero, axis=1)

        return kl_df


class StockBaseMarket(six.with_metaclass(ABCMeta, BaseMarket)):
    """基于股票类型的数据源抽象基类"""

    @abstractmethod
    def minute(self, *args, **kwargs):
        """分钟k线接口"""
        pass

    @abstractmethod
    def kline(self, n_folds=2, start=None, end=None):
        """日k线接口"""
        pass

    @classmethod
    def _fix_kline_pd(cls, kl_df, n_folds, start=None, end=None):
        """修复kline接口的返回金融时间序列"""
        return cls._fix_kline_pd_se(kl_df, n_folds, start=start, end=end)


class FuturesBaseMarket(six.with_metaclass(ABCMeta, BaseMarket)):
    """基于期货类型的数据源抽象基类"""

    @abstractmethod
    def kline(self, n_folds=2, start=None, end=None):
        """日k线接口"""
        pass

    @classmethod
    def _fix_kline_pd(cls, kl_df, n_folds, start=None, end=None):
        """修复kline接口的返回金融时间序列"""
        kl_df = cls._fix_kline_pd_se(kl_df, n_folds, start=start, end=end)
        # 期货数据要修复交割0 close的bar
        return cls._fix_kline_pd_zero(kl_df)


class TCBaseMarket(six.with_metaclass(ABCMeta, BaseMarket)):
    """基于比特币，莱特币等类型的数据源抽象基类"""

    @abstractmethod
    def kline(self, n_folds=2, start=None, end=None):
        """日k线接口"""
        pass

    @abstractmethod
    def minute(self, *args, **kwargs):
        """比特币量化日内短线频繁，需要定制自己的日内策略"""
        pass

    @classmethod
    def _fix_kline_pd(cls, kl_df, n_folds, start=None, end=None):
        """修复kline接口的返回金融时间序列"""
        return cls._fix_kline_pd_se(kl_df, n_folds, start=start, end=end)



from __future__ import absolute_import

from .ABuWGStockInfo import WidgetStockInfo, WidgetSearchStockInfo
from .ABuWGBRunBase import WidgetRunTT
from .ABuWGBSymbol import WidgetSymbolChoice
from .ABuWGBRun import WidgetRunLoopBack
from .ABuWGQuantTool import WidgetQuantTool
from .ABuWGUpdate import WidgetUpdate

__all__ = [
    'WidgetRunLoopBack',
    'WidgetQuantTool',

    'WidgetStockInfo',
    'WidgetSearchStockInfo',

    'WidgetRunTT',
    'WidgetSymbolChoice',
    'WidgetUpdate'
]



from __future__ import absolute_import

# noinspection all
from . import ABuUmpEdgeBase as edge
# noinspection all
from . import ABuUmpMainBase as main
# noinspection all
from . import ABuUmpManager as manager
# noinspection PyUnresolvedReferences
from .ABuUmpBase import ump_main_make_xy, ump_edge_make_xy
# noinspection PyUnresolvedReferences
from .ABuUmpEdgeBase import EEdgeType



# -*- encoding:utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import logging

from ..AlphaBu.ABuPickStockMaster import AbuPickStockMaster
from ..AlphaBu.ABuPickTimeMaster import AbuPickTimeMaster
from ..CoreBu import ABuEnv
from ..CoreBu import ABuStore
from ..CoreBu.ABuEnv import EMarketDataFetchMode
from ..CoreBu.ABuStore import AbuResultTuple
from ..MarketBu.ABuMarket import all_symbol
from ..MarketBu.ABuSymbolPd import kl_df_dict_parallel
from ..TradeBu.ABuBenchmark import AbuBenchmark
from ..TradeBu.ABuCapital import AbuCapital
from ..TradeBu.ABuKLManager import AbuKLManager
from ..UtilBu import ABuDateUtil
from ..UtilBu import ABuProgress

__author__ = '阿布'
__weixin__ = 'abu_quant'


def run_loop_back(read_cash, buy_factors, sell_factors, stock_picks=None, choice_symbols=None, n_folds=2,
                  start=None,
                  end=None,
                  commission_dict=None,
                  n_process_kl=None,
                  n_process_pick=None):
    """
    封装执行择时，选股回测。

    推荐在使用abu.run_loop_back()函数进行全市场回测前使用abu.run_kl_update()函数首先将数据进行更新，
    在run_kl_update()中它会首选强制使用网络数据进行更新，在更新完毕后，更改数据获取方式为本地缓存，
    使用abu.run_kl_update()的好处是将数据更新与策略回测分离，在运行效率及问题排查上都会带来正面的提升

    :param read_cash: 初始化资金额度，eg：1000000
    :param buy_factors: 回测使用的买入因子策略序列，
                    eg：
                        buy_factors = [{'xd': 60, 'class': AbuFactorBuyBreak},
                                       {'xd': 42, 'class': AbuFactorBuyBreak}]
    :param sell_factors: 回测使用的卖出因子序列，
                    eg:
                        sell_factors = [{'stop_loss_n': 0.5, 'stop_win_n': 3.0, 'class': AbuFactorAtrNStop},
                                        {'pre_atr_n': 1.0, 'class': AbuFactorPreAtrNStop},
                                        {'close_atr_n': 1.5, 'class': AbuFactorCloseAtrNStop},]
    :param stock_picks: 回测使用的选股因子序列：
                    eg:
                        stock_pickers = [{'class': AbuPickRegressAngMinMax,
                                          'threshold_ang_min': 0.0, 'reversed': False},
                                         {'class': AbuPickStockPriceMinMax,
                                          'threshold_price_min': 50.0,
                                          'reversed': False}]
    :param choice_symbols: 备选股票池, 默认为None，即使用abupy.env.g_market_target的市场类型进行全市场回测，
                           为None的情况下为symbol序列
                    eg:
                        choice_symbols = ['usNOAH', 'usSFUN', 'usBIDU', 'usAAPL', 'usGOOG',
                                          'usTSLA', 'usWUBA', 'usVIPS']
    :param n_folds: int, 回测n_folds年的历史数据
    :param start: 回测开始的时间, str对象, eg: '2013-07-10'
    :param end: 回测结束的时间, str对象 eg: '2016-07-26'
    :param commission_dict: 透传给AbuCapital，自定义交易手续费的时候时候。
                    eg：
                        def free_commission(trade_cnt, price):
                            # 免手续费
                            return 0
                        commission_dict = {'buy_commission_func': free_commission,
                                         'sell_commission_func': free_commission}
                        AbuCapital(read_cash, benchmark, user_commission_dict=commission_dict)

    :param n_process_kl: 金融时间序列数据收集启动并行的进程数，默认None, 内部根据cpu数量分配
    :param n_process_pick: 择时与选股操作启动并行的进程数，默认None, 内部根据cpu数量分配
    :return: (AbuResultTuple对象, AbuKLManager对象)
    """
    if start is not None and end is not None and ABuDateUtil.date_str_to_int(end) - ABuDateUtil.date_str_to_int(
            start) <= 0:
        logging.info('end date <= start date!!')
        return None, None

    benchmark = AbuBenchmark(n_folds=n_folds, start=start, end=end)
    # 资金类初始化
    capital = AbuCapital(read_cash, benchmark, user_commission_dict=commission_dict)

    """
         win_to_one:
         1. 如果symbol数量少于20
         2. 并且操作系统是windows，因为windows进程开辟销毁开销都非常大，
         3. 判断cpu不是很快，只能通过cpu数量做判断，4核认为速度一般
         这种情况下不再启动多个进程，只使用一个进程运行所有择时选股操作

         TODO：不能只以symbol数量进行判断，结合策略买入卖出策略数进行综合判断
    """
    win_to_one = choice_symbols is not None and len(
        choice_symbols) < 20 and not ABuEnv.g_is_mac_os and ABuEnv.g_cpu_cnt <= 4

    if n_process_pick is None:
        # 择时，选股并行操作的进程等于cpu数量, win_to_one满足情况下1个
        n_process_pick = 1 if win_to_one else ABuEnv.g_cpu_cnt
    if n_process_kl is None:
        # mac系统下金融时间序列数据收集启动两倍进程数, windows只是进程数量，win_to_one满足情况下1个
        n_process_kl = 1 if win_to_one else ABuEnv.g_cpu_cnt * 2 if ABuEnv.g_is_mac_os else ABuEnv.g_cpu_cnt

    # 选股策略执行，多进程方式
    choice_symbols = AbuPickStockMaster.do_pick_stock_with_process(capital, benchmark,
                                                                   stock_picks, choice_symbols=choice_symbols,
                                                                   n_process_pick_stock=n_process_pick)

    if choice_symbols is None or len(choice_symbols) == 0:
        logging.info('pick stock result is zero!')
        return None, None
    # kl数据管理类初始化
    kl_pd_manager = AbuKLManager(benchmark, capital)
    # 批量获取择时kl数据
    kl_pd_manager.batch_get_pick_time_kl_pd(choice_symbols, n_process=n_process_kl)

    # 在择时之前清理一下输出, 不能wait, windows上一些浏览器会卡死
    ABuProgress.do_clear_output(wait=False)

    # 择时策略运行，多进程方式
    orders_pd, action_pd, all_fit_symbols_cnt = AbuPickTimeMaster.do_symbols_with_same_factors_process(
        choice_symbols, benchmark,
        buy_factors, sell_factors, capital, kl_pd_manager=kl_pd_manager, n_process_kl=n_process_kl,
        n_process_pick_time=n_process_pick)

    # 都完事时检测一下还有没有ui进度条
    ABuProgress.do_check_process_is_dead()

    # 返回namedtuple， ('orders_pd', 'action_pd', 'capital', 'benchmark')
    abu_result = AbuResultTuple(orders_pd, action_pd, capital, benchmark)
    # store_abu_result_tuple(abu_result, n_folds)
    return abu_result, kl_pd_manager


def run_kl_update(n_folds=2, start=None, end=None, market=None, n_jobs=16, how='thread'):
    """
    推荐在使用abu.run_loop_back()函数进行全市场回测前使用abu.run_kl_update()函数首先将数据进行更新，
    在run_kl_update()中它会首选强制使用网络数据进行更新，在更新完毕后，更改数据获取方式为本地缓存
    在run_kl_update实现根据EMarketTargetType类型即市场类型，进行全市场金融时间序列数据获取，使用多进
    程或者多线程对外执行函数，多任务批量获取时间序列数据。

    使用abu.run_kl_update()的好处是将数据更新与策略回测分离，在运行效率及问题排查上都会带来正面的提升

    eg：
        from abupy import abu，EMarketTargetType
        # 港股全市场获取
        abupy.env.g_market_target = EMarketTargetType.E_MARKET_TARGET_HK
        # 更新6年的数据
        abu.run_kl_update(n_folds=6)

        # A股全市场获取
        abupy.env.g_market_target = EMarketTargetType.E_MARKET_TARGET_CN
        # 2013-07-10直到2016-07-26的数据
        abu.run_kl_update(start='2013-07-10', end='2016-07-26')

    :param n_folds: 请求几年的历史回测数据int
    :param start: 请求的开始日期 str对象, eg: '2013-07-10'
    :param end: 请求的结束日期 str对象 eg: '2016-07-26'
    :param market: 需要查询的市场，eg：EMarketTargetType.E_MARKET_TARGET_US
    :param n_jobs: 并行的任务数，对于进程代表进程数，线程代表线程数
    :param how: process：多进程，thread：多线程，main：单进程单线程
    """

    pre_market = None
    if market is not None:
        # 临时缓存之前的市场设置
        pre_market = ABuEnv.g_market_target
        ABuEnv.g_market_target = market

    # 所有任务数据强制网络更新
    ABuEnv.g_data_fetch_mode = EMarketDataFetchMode.E_DATA_FETCH_FORCE_NET
    # index=True, 需要大盘数据
    symbols = all_symbol(index=True)
    _ = kl_df_dict_parallel(symbols, n_folds=n_folds, start=start, end=end, n_jobs=n_jobs, how=how)
    # 完成更新后所有认为强制走本地数据
    ABuEnv.g_data_fetch_mode = EMarketDataFetchMode.E_DATA_FETCH_FORCE_LOCAL

    if market is not None and pre_market is not None:
        # 还原缓存的市场设置
        ABuEnv.g_market_target = pre_market


def store_abu_result_tuple(abu_result_tuple, n_folds, store_type=None, custom_name=None):
    """
    保存abu.run_loop_back的回测结果AbuResultTuple对象，根据n_folds，store_type参数
    来定义存储的文件名称，透传参数使用ABuStore.store_abu_result_tuple执行操作

    :param abu_result_tuple: AbuResultTuple对象类型
    :param n_folds: 回测执行了几年，只影响存贮文件名
    :param store_type: 回测保存类型EStoreAbu类型，只影响存贮文件名
    :param custom_name: 如果store_type=EStoreAbu.E_STORE_CUSTOM_NAME时需要的自定义文件名称
    """
    ABuStore.store_abu_result_tuple(abu_result_tuple, n_folds, store_type=store_type, custom_name=custom_name)


def load_abu_result_tuple(n_folds, store_type, custom_name=None):
    """
    读取使用store_abu_result_tuple保存的回测结果，根据n_folds，store_type参数
    来定义读取的文件名称，依次读取orders_pd，action_pd，capital，benchmark后构造
    AbuResultTuple对象返回，透传参数使用ABuStore.load_abu_result_tuple执行操作

    :param n_folds: 回测执行了几年，只影响读取的文件名
    :param store_type: 回测保存类型EStoreAbu类型，只影响读取的文件名
    :param custom_name: 如果store_type=EStoreAbu.E_STORE_CUSTOM_NAME时需要的自定义文件名称
    :return: AbuResultTuple对象
    """
    return ABuStore.load_abu_result_tuple(n_folds, store_type, custom_name=custom_name)


# noinspection PyUnusedLocal
def gen_buy_from_chinese(*args, **kwargs):
    """
    抱歉！由于中文生成策略的方法也需要遵循一定的语法和句式，对于完全不熟悉编程的人可能会产生错误，'
          '造成无谓的经济损失，所以中文自动生成交易策略模块暂时不开放接口以及源代码！
    """

    print('抱歉！由于中文生成策略的方法也需要遵循一定的语法和句式，对于完全不熟悉编程的人可能会产生错误，'
          '造成无谓的经济损失，所以中文自动生成交易策略模块暂时不开放接口以及源代码！')



# -*- encoding:utf-8 -*-
"""示例ump主裁特征跳空模块"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from ..TradeBu.ABuMLFeature import AbuFeatureJump
from ..MLBu.ABuMLPd import AbuMLPd
from .ABuUmpBase import ump_main_make_xy, BuyUmpMixin
from .ABuUmpMainBase import AbuUmpMainBase

__author__ = '阿布'
__weixin__ = 'abu_quant'


class AbuUmpMainJump(AbuUmpMainBase, BuyUmpMixin):
    """主裁跳空特征类，AbuUmpMainBase子类，混入BuyUmpMixin，做为买入ump类"""

    class UmpJumpFiter(AbuMLPd):
        """
            内部类，只需要实现make_xy，且使用ump_main_make_xy装饰

            在__init__中通过：
                    self.fiter_cls = self.get_fiter_class()
                    self.fiter = self.fiter_cls(orders_pd=self.orders_pd, **kwarg)
            构造裁判的filter以及重要的self.fiter.df，即pd.DataFrame对象特征
        """

        @ump_main_make_xy
        def make_xy(self, **kwarg):
            """
            make_xy通过装饰器ump_main_make_xy进行二次包装
            这里只需要使用filter选取需要的特征，即从self.order_has_ret中使用filter选取特征列形成df
            :param kwarg: ump_main_make_xy装饰器中使用kwarg
                          kwargs['orders_pd'] 做为必须要有的关键字参数：交易训练集数据，pd.DataFrame对象
                          kwargs['scaler']    做为可选关键字参数：控制在make_xy中返回的特征矩阵数据是否进行标准化处理

            :return: self.order_has_ret中使用filter选取特征列形成jump_df
                     ump_main_make_xy装饰器在make_xy返回jump_df后做转换matrix，形成x，y等工作
            """
            # 'regex=result|buy_diff_down_days|buy_diff_up_days|buy_jump_down_power|buy_jump_up_power'
            regex = 'result|{}'.format('|'.join(AbuFeatureJump().get_feature_ump_keys(ump_cls=AbuUmpMainJump)))
            # noinspection PyUnresolvedReferences
            jump_df = self.order_has_ret.filter(regex=regex)
            """
                eg: jump_df
                            result  buy_jump_down_power  buy_diff_down_days  \
                2014-09-24       0                0.000                   0
                2014-10-24       0                0.000                   0
                2014-10-29       1               -1.109                 278
                2014-10-29       1                0.000                   0
                2014-10-29       1               -1.522                  75
                2014-10-29       1               -1.109                 278
                2014-11-03       1               -1.451                  78
                2014-11-11       1                0.000                   0
                2014-11-12       0               -1.863                  82
                2014-11-26       0               -2.115                  96
                ...            ...                  ...                 ...
                2016-03-14       1               -2.965                  56
                2016-03-14       0               -2.323                  67
                2016-03-30       0               -1.744                  74
                2016-04-04       1               -2.085                  36
                2016-04-13       0               -2.940                  47
                2016-04-14       0               -1.764                 100
                2016-04-15       0               -1.455                 101
                2016-04-15       0               -1.455                 101
                2016-04-29       1               -1.178                  24
                2016-04-29       1               -1.178                  24

                            buy_jump_up_power  buy_diff_up_days
                2014-09-24              3.344                61
                2014-10-24              3.344                91
                2014-10-29              2.920                95
                2014-10-29              1.283               243
                2014-10-29              3.727               238
                2014-10-29              2.920                95
                2014-11-03              1.014               214
                2014-11-11              1.291               256
                2014-11-12              1.606               256
                2014-11-26              1.772               270
                ...                       ...               ...
                2016-03-14              2.682                14
                2016-03-14              2.478               246
                2016-03-30              2.621                32
                2016-04-04              1.507                45
                2016-04-13              1.678                56
                2016-04-14              1.158                71
                2016-04-15              1.075                58
                2016-04-15              1.075                58
                2016-04-29              3.259                 0
                2016-04-29              3.259                 0
            """
            return jump_df

    def get_predict_col(self):
        """
        主裁跳空特征keys：['buy_diff_down_days', 'buy_diff_up_days', 'buy_jump_down_power', 'buy_jump_up_power']
        :return: ['buy_diff_down_days', 'buy_diff_up_days', 'buy_jump_down_power', 'buy_jump_up_power']
        """
        return AbuFeatureJump().get_feature_ump_keys(ump_cls=AbuUmpMainJump)

    def get_fiter_class(self):
        """
        主裁特征跳空返回的AbuMLPd子类：AbuUmpMainJump.UmpJumpFiter
        :return: AbuUmpMainJump.UmpJumpFiter
        """
        return AbuUmpMainJump.UmpJumpFiter

    @classmethod
    def class_unique_id(cls):
        """
        具体ump类关键字唯一名称，类方法：return 'jump_main'
        主要针对外部user设置自定义ump使用, 需要user自己保证class_unique_id的唯一性，内部不做检测
        具体使用见ABuUmpManager中extend_ump_block方法
        """
        return 'jump_main'



# -*- encoding:utf-8 -*-
"""
    交易订单模块
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import math

import numpy as np

from ..CoreBu import ABuEnv
from ..CoreBu.ABuEnv import EMarketTargetType
from ..BetaBu import ABuPositionBase
from ..MarketBu.ABuSymbolFutures import AbuFuturesCn, AbuFuturesGB
from ..MarketBu.ABuHkUnit import AbuHkUnit
from ..MarketBu import ABuMarket
from ..MarketBu.ABuMarket import MarketMixin

__author__ = '阿布'
__weixin__ = 'abu_quant'


class OrderMarket(MarketMixin):
    """根据AbuOrder对象，设置混入类MarketMixin中symbol_name"""

    def __init__(self, order):
        """
        设置混入类MarketMixin中symbol_name，以获取symbol_market等信息
        :param order: AbuOrder对象
        """
        if isinstance(order, AbuOrder):
            self.symbol_name = order.buy_symbol
        else:
            raise TypeError('order is AbuOrder object!!!')


# noinspection PyAttributeOutsideInit
class AbuOrder(object):
    """交易订单类"""

    # 多个因子买入条件可能生成几百万个order对象使用__slots__降低内存消耗
    __slots__ = ('order_deal', 'buy_symbol', 'buy_date', 'buy_factor', 'buy_factor_class',
                 'buy_price', 'buy_cnt', 'buy_pos', 'sell_date',
                 'buy_type_str', 'expect_direction',
                 'sell_type', 'keep_days', 'sell_price', 'sell_type_extra', 'ml_features')

    def __init__(self):
        """初始设置只需要将order_deal设置未成交状态"""
        self.order_deal = False

    def fit_buy_order(self, day_ind, factor_object):
        """
        根据买入交易日当当天数据以及买入因子，拟合计算买入订单
        :param day_ind: 买入交易发生的时间索引，即对应self.kl_pd.key
        :param factor_object: ABuFactorBuyBases子类实例对象
        """
        kl_pd = factor_object.kl_pd
        # 要执行买入当天的数据
        kl_pd_buy = kl_pd.iloc[day_ind + 1]
        # 买入因子名称
        factor_name = factor_object.factor_name if hasattr(factor_object, 'factor_name') else 'unknown'
        # 日内滑点决策类
        slippage_class = factor_object.slippage_class
        # 仓位管理类设置
        position_class = factor_object.position_class
        # 初始资金，也可修改策略使用剩余资金
        read_cash = factor_object.capital.read_cash
        # 实例化滑点类
        fact = slippage_class(kl_pd_buy, factor_name)
        # 执行fit_price, 计算决策买入价格
        bp = fact.fit()
        # 如果滑点类中决定不买入，撤单子，bp就返回正无穷
        if bp < np.inf:
            # 默认保证金比例是1，即没有杠杆，修改ABuPositionBase.g_deposit_rate可提高融资能力，
            # 如果margin＝2－>ABuPositionBase.g_deposit_rate = 0.5, 即只需要一半的保证金
            deposit_rate = ABuPositionBase.g_deposit_rate
            # 实例化仓位管理类
            position = position_class(kl_pd_buy, factor_name, factor_object.kl_pd.name, bp, read_cash, deposit_rate)

            market = ABuEnv.g_market_target if ABuMarket.g_use_env_market_set else position.symbol_market

            """
                由于模块牵扯复杂，暂时不迁移保证金融资相关模块，期货不使用杠杆，即回测不牵扯资金总量的评估
                if market == EMarketTargetType.E_MARKET_TARGET_FUTURES_CN:
                    deposit_rate = 0.10
                    q_df = AbuFuturesCn().query_symbol(factor_object.kl_pd.name)
                    if q_df is not None:
                        deposit_rate = q_df.min_deposit.values[0]
                    # 重新设置保证金比例
                    position.deposit_rate = deposit_rate
            """
            # 执行fit_position，通过仓位管理计算买入的数量
            bc = position.fit_position(factor_object)
            if np.isnan(bc):
                return

            if market != EMarketTargetType.E_MARKET_TARGET_TC:
                # 除了比特币市场外，都向下取整数到最小交易单位个数
                buy_cnt = int(math.floor(bc))
            else:
                # 币类市场可以买非整数个, 保留三位小数
                buy_cnt = round(bc, 3)

            if market == EMarketTargetType.E_MARKET_TARGET_US:
                # 美股1
                min_cnt = 1
            elif market == EMarketTargetType.E_MARKET_TARGET_TC:
                # 国内一般只支持到0.01个
                min_cnt = 0.01
            elif market == EMarketTargetType.E_MARKET_TARGET_CN:
                # A股最小100一手
                min_cnt = 100
                # 向最小的手量看齐
                buy_cnt -= buy_cnt % min_cnt
            elif market == EMarketTargetType.E_MARKET_TARGET_HK:
                # 港股从AbuHkUnit读取数据，查询对应symbol每一手的交易数量
                min_cnt = AbuHkUnit().query_unit(factor_object.kl_pd.name)
                # 向最小的手量看齐
                buy_cnt -= buy_cnt % min_cnt
            elif market == EMarketTargetType.E_MARKET_TARGET_FUTURES_CN:
                # 国内期货，查询最少一手单位
                min_cnt = AbuFuturesCn().query_min_unit(factor_object.kl_pd.name)
                # 向最小的手量看齐
                buy_cnt -= buy_cnt % min_cnt
            elif market == EMarketTargetType.E_MARKET_TARGET_OPTIONS_US:
                # 美股期权最小合约单位1contract，代表100股股票权利
                min_cnt = 100
                buy_cnt -= buy_cnt % min_cnt
            elif market == EMarketTargetType.E_MARKET_TARGET_FUTURES_GLOBAL:
                # 国际期货, 查询最少一手单位
                min_cnt = AbuFuturesGB().query_min_unit(factor_object.kl_pd.name)
                buy_cnt -= buy_cnt % min_cnt
            else:
                raise TypeError('ABuEnv.g_market_target ERROR, market={}, g_market_target={}'.format(
                    market, ABuEnv.g_market_target))

            if buy_cnt < min_cnt:
                # 不够买最少单位量
                return

            # 如下生成order内部数据
            self.buy_symbol = kl_pd.name
            # 订单写入买入日期
            self.buy_date = int(kl_pd_buy.date)
            # 订单写入买入因子名字
            self.buy_factor = factor_name
            # 订单对象中添加买入因子类名，和buy_factor不同没有具体参数等唯一key
            self.buy_factor_class = factor_object.__class__.__name__
            # 订单写入买入价格
            self.buy_price = bp
            # 订单写入买入数量
            self.buy_cnt = buy_cnt
            # 订单写入仓位管理类名称
            self.buy_pos = position.__class__.__name__
            # 订单写入买入类型，call or put
            self.buy_type_str = factor_object.buy_type_str
            # 订单写入买入因子期望方向
            self.expect_direction = factor_object.expect_direction

            # 如下卖出信息具体写入在fit_sell_order中
            # 订单卖出时间
            self.sell_date = None
            # 订单卖出类型，keep：持有
            self.sell_type = 'keep'
            # 交易日持有天数
            self.keep_days = 0

            # 订单卖出价格
            self.sell_price = None
            # 订单卖出额外信息
            self.sell_type_extra = ''

            # 订单买入，卖出特征
            self.ml_features = None
            # 订单形成
            self.order_deal = True

    def fit_sell_order(self, day_ind, factor_object):
        """
        根据卖出交易日当当天数据以及卖出因子，拟合计算卖出信息，完成订单
        :param day_ind: 卖出交易发生的时间索引，即对应self.kl_pd.key
        :param factor_object: AbuFactorSellBase子类实例对象
        """

        if self.sell_type != 'keep':
            # 保证外部不需要过滤单子，内部自己过滤已经卖出成交的订单
            return

        # 卖出策略中进行特征合成, 以及ump拦截卖出行为
        if factor_object.make_sell_order(self, day_ind):
            kl_pd_sell = factor_object.kl_pd.iloc[day_ind + 1]

            # 日内滑点决策类
            slippage_class = factor_object.slippage_class
            # 卖出因子名称
            factor_name = factor_object.factor_name if hasattr(factor_object, 'factor_name') else 'unknown'
            # 实例化日内滑点决策类，进行具体卖出价格决策
            sell_price = slippage_class(kl_pd_sell, factor_name).fit()

            if sell_price == -np.inf:
                # 如果卖出执行返回负无穷说明无法卖出，例如跌停
                return

            self.sell_price = sell_price
            # 卖出原因其它描述
            sell_type_extra = factor_object.sell_type_extra if hasattr(factor_object, 'sell_type_extra') else 'unknown'
            self.sell_type_extra = sell_type_extra
            if self.buy_type_str == 'call':
                # call卖出类型:  win = self.sell_price > self.buy_price
                self.sell_type = 'win' if self.sell_price > self.buy_price else 'loss'
            else:
                # put卖出类型:  loss = self.sell_price > self.buy_price
                self.sell_type = 'loss' if self.sell_price > self.buy_price else 'win'
            # 卖出日期写入单子
            self.sell_date = int(kl_pd_sell.date)

    def __str__(self):
        """打印对象显示：buy_symbol， buy_price， buy_cnt， buy_date，buy_factor，sell_date，sell_type， sell_price"""
        return 'buy Symbol = ' + str(self.buy_symbol) + '\n' \
               + 'buy Prices = ' + str(self.buy_price) + '\n' \
               + 'buy cnt = ' + str(self.buy_cnt) + '\n' \
               + 'buy date = ' + str(self.buy_date) + '\n' \
               + 'buy factor = ' + str(self.buy_factor) + '\n' \
               + 'sell date = ' + str(self.sell_date) + '\n' \
               + 'sell type = ' + str(self.sell_type) + '\n' \
               + 'sell Price = ' + str(self.sell_price) + '\n'

    __repr__ = __str__



from __future__ import absolute_import

from .ABuDataBase import BaseMarket, FuturesBaseMarket, StockBaseMarket, TCBaseMarket, SupportMixin
from .ABuDataParser import AbuDataParseWrap
from . import ABuSymbolPd
from .ABuSymbolPd import get_price
from .ABuSymbol import IndexSymbol, Symbol, code_to_symbol, search_to_symbol_dict
from . import ABuSymbol
from ..MarketBu.ABuSymbolStock import AbuSymbolCN, AbuSymbolUS, AbuSymbolHK, query_stock_info
from .ABuSymbolFutures import AbuFuturesCn, AbuFuturesGB
from .ABuHkUnit import AbuHkUnit
from . import ABuMarket
from .ABuMarket import MarketMixin
from . import ABuIndustries
from . import ABuMarketDrawing
from . import ABuNetWork

__all__ = [
    'BaseMarket',
    'FuturesBaseMarket',
    'StockBaseMarket',
    'TCBaseMarket',
    'SupportMixin',
    'AbuDataParseWrap',
    'MarketMixin',
    'ABuSymbolPd',
    'get_price',
    'ABuSymbol',
    'AbuSymbolCN',
    'AbuFuturesGB',
    'AbuSymbolUS',
    'AbuSymbolHK',
    'query_stock_info',
    'AbuFuturesCn',
    'AbuHkUnit',
    'ABuMarket',
    'IndexSymbol',
    'Symbol',
    'code_to_symbol',
    'search_to_symbol_dict',
    'ABuIndustries',
    'ABuMarketDrawing',
    'ABuNetWork'
]



# -*- encoding:utf-8 -*-
"""
    选股因子抽象基类
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import functools
from abc import ABCMeta, abstractmethod

from ..CoreBu.ABuFixes import six
from ..CoreBu import ABuEnv
from ..CoreBu.ABuBase import AbuParamBase

__author__ = '阿布'
__weixin__ = 'abu_quant'


def reversed_result(func):
    """对选股结果进行反转的装饰器，装饰在fit_pick上"""
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        result = func(self, *args, **kwargs)
        # 通过对象变量reversed，判断是否反转结果
        result = not result if self.reversed else result
        return result

    return wrapper


class AbuPickStockBase(six.with_metaclass(ABCMeta, AbuParamBase)):
    def __init__(self, capital, benchmark, **kwargs):
        """
        :param capital:资金类AbuCapital实例化对象
        :param benchmark:交易基准对象，AbuBenchmark实例对象
        :param kwargs:其它可扩展参数
        """
        self.capital = capital
        self.benchmark = benchmark

        # 所有自定义参数不使用kwargs.pop('a', default)方式，因为有从配置文件读取等需求，而且后续_init_self可能也还需要
        # 默认反转结果false，通过kwargs参数控制
        self.reversed = False
        if 'reversed' in kwargs:
            self.reversed = kwargs['reversed']

        # 默认选股周期默认一年的交易日
        self.xd = ABuEnv.g_market_trade_year
        if 'xd' in kwargs:
            self.xd = kwargs['xd']

        # 最小选股周期，小于这个将抛弃，即结果投反对票
        self.min_xd = int(self.xd / 2)
        if 'min_xd' in kwargs:
            self.min_xd = kwargs['min_xd']

        # 因子独有的init继续
        self._init_self(**kwargs)

    def __str__(self):
        """打印对象显示：class name, benchmark, reversed, xd, min_xd"""
        return '{}: {}, reversed:{}, xd:{}, min_xd:{}'.format(self.__class__.__name__,
                                                              self.benchmark, self.reversed, self.xd, self.min_xd)

    __repr__ = __str__

    @abstractmethod
    def _init_self(self, **kwargs):
        """子类因子针对可扩展参数的初始化"""
        pass

    @abstractmethod
    def fit_pick(self, *args, **kwargs):
        """选股操作接口，即因子对象针对一个交易目标的投票结果，具体详见示例因子"""
        pass

    @abstractmethod
    def fit_first_choice(self, pick_worker, choice_symbols, *args, **kwargs):
        """因子首选批量选股接口，即因子对象对多个交易目标的投票结果，具体详见示例因子"""
        pass



# -*- encoding:utf-8 -*-
"""
    行业分类模块，仅支持美股，a股，港股
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from collections import Iterable
import logging
from fnmatch import fnmatch

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from ..MarketBu import ABuSymbolPd
from ..UtilBu.ABuStrUtil import to_unicode
from ..UtilBu import ABuScalerUtil
from ..MarketBu.ABuSymbolStock import AbuSymbolUS, AbuSymbolCN, AbuSymbolHK
from ..CoreBu.ABuEnv import EMarketDataSplitMode, EMarketTargetType
from ..CoreBu import ABuEnv


def industries_df(target_symbol):
    """
    分别查询target_symbol是否在美股，a股，港股中有对应的行业分类，如果查询到
    返回查询的结果industries（pd.DataFrame对象）以及 所属symbol市场类，即
    AbuSymbolStockBase子类
    :param target_symbol: symbol str对象
    :return: （返回查询的结果industries: pd.DataFrame对象, 所属symbol市场类: AbuSymbolStockBase子类）
    """
    industries = AbuSymbolUS().query_industry_symbols(target_symbol)
    if industries is not None:
        # 美股查到了有行业分类结果
        return industries, AbuSymbolUS()

    industries = AbuSymbolCN().query_industry_symbols(target_symbol)
    if industries is not None:
        # a股查到了有行业分类结果
        return industries, AbuSymbolCN()

    industries = AbuSymbolHK().query_industry_symbols(target_symbol)
    if industries is not None:
        # 港股查到了有行业分类结果
        return industries, AbuSymbolHK()
    # 都没查到，返回None, None
    return None, None


def industries_factorize(market=None):
    """
    查询market所在市场的行业分类离散值，默认market=None，即使用ABuEnv.g_market_target
    :param market: 需要查询的市场，eg：EMarketTargetType.E_MARKET_TARGET_US
    :return: 对应market的行业分类简述和factorize值，pd.Series对象
            eg：
                1      中国食品、饮料与烟草
                2        油气/钻孔与探测
                3          出版业/报纸
                4            生物技术
                5          数据存储设备
                6            教育培训
                7              电视
                8          中国网络游戏
                9           中国新能源
                          ...
                249       食品多样化经营
                250       电影制作及影院
                251           农产品
                252       烟草制品及其它
                253             铝
                254       汽车配件批发商
                255         家具装饰店
                256          乳酪产品
                257          医药生物
                258        中国酒店餐饮
           eg：查询A股市场行业：
                input：
                        ABuIndustries.industries_factorize(market=EMarketTargetType.E_MARKET_TARGET_CN)
                output：
                        1     商业贸易
                        2     有色金属
                        3     电气设备
                        4     家用电器
                        5     建筑装饰
                        6      计算机
                        7     轻工制造
                        8     机械设备
                        9     医药生物
                              ...
                        25    国防军工
                        26    A股指数
                        27    非银金融
                        28    建筑建材
                        29      银行
                        30    信息设备
                        31      钢铁
                        32    交运设备
                        33    餐饮旅游
                        34    黑色金属
    """
    if market is None:
        # None则服从ABuEnv.g_market_target市场设置
        market = ABuEnv.g_market_target

    if market == EMarketTargetType.E_MARKET_TARGET_US:
        return AbuSymbolUS().industry_factorize_name_series
    elif market == EMarketTargetType.E_MARKET_TARGET_CN:
        return AbuSymbolCN().industry_factorize_name_series
    elif market == EMarketTargetType.E_MARKET_TARGET_HK:
        return AbuSymbolHK().industry_factorize_name_series
    # 仅支持美股，a股，港股
    raise TypeError('JUST SUPPORT US, CN, HK!')


def industries_market(market=None):
    """
    查询market所在市场的句柄对象，即：
        美股市场E_MARKET_TARGET_US：AbuSymbolUS对象
        a股市场E_MARKET_TARGET_CN：AbuSymbolCN对象
        港股市场E_MARKET_TARGET_HK：AbuSymbolHK对象
    :param market: 需要查询的市场，eg：EMarketTargetType.E_MARKET_TARGET_US
    :return: AbuSymbolUS对象 or AbuSymbolCN对象 or AbuSymbolHK对象
    """
    if market is None:
        # None则服从ABuEnv.g_market_target市场设置
        market = ABuEnv.g_market_target

    if market == EMarketTargetType.E_MARKET_TARGET_US:
        industries_market_op = AbuSymbolUS()
    elif market == EMarketTargetType.E_MARKET_TARGET_CN:
        industries_market_op = AbuSymbolCN()
    elif market == EMarketTargetType.E_MARKET_TARGET_HK:
        industries_market_op = AbuSymbolHK()
    else:
        raise TypeError('JUST SUPPORT US, CN, HK!')
    # 返回市场的句柄操作对象
    return industries_market_op


def match_industries_factorize(match, market=None):
    """
        通过模糊查询market所在市场中match关键字所示的所有行业factorize，
        获取factorize后可通过query_factorize_industry_df将factorize
        所示行业进行获取
        eg：
        input：
                ABuIndustries.match_industries_factorize('中国')
        output：
                [(1, '中国食品、饮料与烟草'),
                 (8, '中国网络游戏'),
                 (9, '中国新能源'),
                 (22, '中国汽车与汽车零部件'),
                 (31, '中国制药、生物科技和生命科学'),
                 (32, '中国金融'),
                 (33, '中国互联网软件服务'),
                 (41, '中国金属与采矿'),
                 (54, '中国建筑材料'),
                 (66, '中国硬件、半导体与设备'),
                 (79, '中国运输'),
                 (81, '中国化学制品'),
                 (114, '中国互联网信息服务'),
                 (169, '中国房地产'),
                 (195, '中国电子商务'),
                 (212, '中国耐用消费品与服装'),
                 (214, '中国一般制造业'),
                 (216, '中国媒体'),
                 (217, '中国日消品零售'),
                 (220, '中国软件与服务'),
                 (223, '中国传统能源'),
                 (224, '中国能源设备与服务'),
                 (228, '中国纸业与包装'),
                 (232, '中国商业与专业服务'),
                 (237, '中国教育培训'),
                 (238, '中国医疗保健设备与服务'),
                 (240, '中国非日消品零售'),
                 (258, '中国酒店餐饮')]
    :param match: 匹配的行业关键字，支持通配符，eg：'医药*'， '*互联网*', '*科技'
    :param market: 需要查询的市场，eg：EMarketTargetType.E_MARKET_TARGET_US
    :return: 匹配的list序列对象，序列中每一个元素为(factorize, 本地描述)，eg：(33, '中国互联网软件服务')
    """
    match = to_unicode(match)
    _industries_factorize = industries_factorize(market=market)
    if u'*' not in match:
        # 如果不带＊那就前后加＊进行match
        match = u'*{}*'.format(match)
    match_list = list()
    for factorize, industries in enumerate(_industries_factorize):
        if fnmatch(to_unicode(industries), match):
            # 使用fnmatch进行匹配
            match_list.append((factorize, industries))

    match = match.replace(u'*', u'')
    if len(match_list) == 0 and len(match) > 1:
        # 如果第一次模糊查询没有查到，开始一个字一个字的进行模糊查询，eg：*教育*－>*教*－>*育*
        for match_pos in np.arange(0, len(match)):
            # 伪递归，迭代递归方法里面由于match = match.replace('*', '')所以不满足迭代再次递归的条件：len(match) > 1
            match_list = match_industries_factorize(match[match_pos], market=market)
            if len(match_list) > 0:
                # 一旦其中一个字的子模糊查询查到了，就直接返回eg：*教*
                break

    # TODO: 如果还没有查到，添加按照拼音等模糊查询匹配方式
    return match_list


def query_match_industries_df(match, market=None):
    """
    通过模糊查询market所在市场中match关键字所示的所有行业信息组装成pd.DataFrame对象
    :param match: 匹配的行业关键字，支持通配符，eg：'医药*'， '*互联网*', '*科技'
    :param market: 需要查询的市场，eg：EMarketTargetType.E_MARKET_TARGET_US
    :return: 返回行业组合pd.DataFrame对象
    """
    # 获取模糊查询factorize序列
    match_list = match_industries_factorize(match, market=market)
    # 通过industries_market获取对应市场操作句柄industries_market_op
    industries_market_op = industries_market(market=market)

    industry_df = None
    for factorize, _ in match_list:
        # 将所有查询到的行业信息，pd.DtaFrame对象连接起来
        query_industry = industries_market_op.query_industry_factorize(factorize)
        industry_df = query_industry if industry_df is None else pd.concat([query_industry, industry_df])
    if industry_df is not None:
        # 去除重复的，比如a在行业b，又在行业c
        # noinspection PyUnresolvedReferences
        industry_df.drop_duplicates(inplace=True)

    return industry_df


def query_factorize_industry_df(factorize_arr, market=None):
    """
    使用match_industries_factorize可以查询到行业所对应的factorize序列，
    使用factorize序列即组成需要查询的行业组合，返回行业组合pd.DataFrame对象
    eg: 从美股所有行业中找到中国企业的行业
        input：ABuIndustries.match_industries_factorize('中国', market=EMarketTargetType.E_MARKET_TARGET_US)
        output：
            [(1, '中国食品、饮料与烟草'),
             (8, '中国网络游戏'),
             (9, '中国新能源'),
             (22, '中国汽车与汽车零部件'),
             (31, '中国制药、生物科技和生命科学'),
             (32, '中国金融'),
             (33, '中国互联网软件服务'),
             (41, '中国金属与采矿'),
             (54, '中国建筑材料'),
             (66, '中国硬件、半导体与设备'),
             (79, '中国运输'),
             (81, '中国化学制品'),
             (114, '中国互联网信息服务'),
             (169, '中国房地产'),
             (195, '中国电子商务'),
             (212, '中国耐用消费品与服装'),
             (214, '中国一般制造业'),
             (216, '中国媒体'),
             (217, '中国日消品零售'),
             (220, '中国软件与服务'),
             (223, '中国传统能源'),
             (224, '中国能源设备与服务'),
             (228, '中国纸业与包装'),
             (232, '中国商业与专业服务'),
             (237, '中国教育培训'),
             (238, '中国医疗保健设备与服务'),
             (240, '中国非日消品零售'),
             (258, '中国酒店餐饮')]

        然后使用ABuIndustries.query_factorize_industry_df((31, 32, 33))即可获取到
             (31, '中国制药、生物科技和生命科学'),
             (32, '中国金融'),
             (33, '中国互联网软件服务'),
        行业中的所有股票信息的pd.DataFrame对象

    :param factorize_arr: eg：(31, 32, 33) or [31, 32, 33] or 31
    :param market: 需要查询的市场，eg：EMarketTargetType.E_MARKET_TARGET_US
    :return: 返回行业组合pd.DataFrame对象
    """
    if not isinstance(factorize_arr, Iterable):
        # 如果不是可迭代的，即只是一个factorize序列，转换为序列，方便统一处理
        factorize_arr = [factorize_arr]
    # 通过industries_market获取对应市场操作句柄industries_market_op
    industries_market_op = industries_market(market=market)

    industry_df = None
    for ind in factorize_arr:
        query_industry = industries_market_op.query_industry_factorize(ind)
        # 将所有查询到的行业信息，pd.DtaFrame对象连接起来
        industry_df = query_industry if industry_df is None else pd.concat([query_industry, industry_df])
    if industry_df is not None:
        # 去除重复的，比如a在行业b，又在行业c
        # noinspection PyUnresolvedReferences
        industry_df.drop_duplicates(inplace=True)

    return industry_df


def query_factorize_industry_symbol(factorize, market=None):
    """
    套接query_factorize_industry_df方法，只返回df在的symbol序列：
        query_factorize_industry_df(factorize, market=market).symbol
    :param factorize: factorize_arr: eg：(31, 32, 33) or [31, 32, 33] or 31
    :param market: 需要查询的市场，eg：EMarketTargetType.E_MARKET_TARGET_US
    :return: pd.Series对象
            eg：
                1. 首先使用industries_factorize查询到outputa股行业分类
                input：
                    ABuIndustries.industries_factorize(market=EMarketTargetType.E_MARKET_TARGET_CN)
                output：
                1     商业贸易
                2     有色金属
                3     电气设备
                4     家用电器
                5     建筑装饰
                6      计算机
                7     轻工制造
                8     机械设备
                9     医药生物
                      ...
                29      银行
                30    信息设备
                31      钢铁
                32    交运设备
                33    餐饮旅游
                34    黑色金属

                2. 使用query_factorize_industry_symbol((6, 9, 29))，查询计算机，医药生物，银行行业symbool
                input：
                    ABuIndustries.query_factorize_industry_symbol((6, 9, 29),
                                  market=EMarketTargetType.E_MARKET_TARGET_CN)
                output：
                    ['sh601939',
                     'sh601398',
                     'sh601288',
                     'sh601328',
                     'sh601009',
                     ..........
                     ..........
                     'sz300513',
                     'sz002236',
                     'sz300044',
                     'sz300302',
                     'sh600756']
    """
    factorize_df = query_factorize_industry_df(factorize, market=market)
    if factorize_df is not None:
        # 通过industries_market获取对应市场操作句柄industries_market_op
        industries_market_op = industries_market(market=market)
        # 通过对应的市场对象op自己去从factorize_df中组装symbol序列
        return industries_market_op.symbol_func(factorize_df)
    # factorize没有匹配上直接返回空序列
    return list()


def query_match_industries_symbol(match, market=None):
    """
    套接query_match_industries_df方法，只返回df在的symbol序列：
        query_match_industries_df(factorize, market=market).symbol
    :param match: 匹配的行业关键字，支持通配符，eg：'医药*'， '*互联网*', '*科技'
    :param market: 需要查询的市场，eg：EMarketTargetType.E_MARKET_TARGET_US
    :return:
            eg：
                input：ABuIndustries.query_match_industries_symbol('医疗', market=EMarketTargetType.E_MARKET_TARGET_CN)
                output：
                        ['sz002826',
                         'sh600645',
                         'sz300534',
                         'sh600080',
                         'sz300595',
                         ..........
                         ..........
                         'sh600518',
                         'sh900904',
                         'sh600993',
                         'sh600332',
                         'sz002589']
    """
    match_df = query_match_industries_df(match, market=market)
    if match_df is not None:
        # 通过industries_market获取对应市场操作句柄industries_market_op
        industries_market_op = industries_market(market=market)
        # 通过对应的市场对象op自己去从factorize_df中组装symbol序列
        return industries_market_op.symbol_func(match_df)
    # match没有匹配上直接返回空序列
    return list()


def get_industries_panel_from_target(target_symbol, show=False, n_folds=2):
    """
    获取target_symbol所在的行业分类中所有的金融时间序列，组成三维pd.Panel对象
    :param target_symbol: symbol str对象
    :param show: 是否可视化行业分类中所有的金融时间序列
    :param n_folds: 获取n_folds年历史交易数据
    :return: (pd.Panel对象, p_date.swapaxes('items', 'minor'))
    """
    df, s_obj = industries_df(target_symbol)
    if df is not None:
        # 通过symbol_func转换为可直接使用ABuSymbolPd.make_kl_df请求的target_symbols序列
        target_symbols = s_obj.symbol_func(df)

        from ..TradeBu.ABuBenchmark import AbuBenchmark
        # 以target_symbol做为标尺实例化benchmark对象为make_kl_df做准备
        benchmark = AbuBenchmark(target_symbol, n_folds=n_folds)
        # 传递序列target_symbols，返回的p_date是pd.Panel对象
        p_date = ABuSymbolPd.make_kl_df(target_symbols, data_mode=EMarketDataSplitMode.E_DATA_SPLIT_UNDO,
                                        benchmark=benchmark, n_folds=n_folds)
        # 内部做个转轴处理，方便外部使用
        p_data_it = p_date.swapaxes('items', 'minor')
        if show:
            # 可视化行业分类中所有的金融时间序列
            data = p_data_it['close']
            data = ABuScalerUtil.scaler_std(data)
            data.plot()
            plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
            plt.ylabel('Price')
            plt.xlabel('Time')
        return p_date, p_data_it
    else:
        logging.info('Industries targetSymbols len = 0')
        return None, None



# -*- encoding:utf-8 -*-
"""示例ump边裁特征走势拟合角度模块"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from ..TradeBu.ABuMLFeature import AbuFeatureDeg, AbuFeatureDegExtend
from ..MLBu.ABuMLPd import AbuMLPd
from .ABuUmpBase import ump_edge_make_xy, BuyUmpMixin
from .ABuUmpEdgeBase import AbuUmpEdgeBase

__author__ = '阿布'
__weixin__ = 'abu_quant'


class AbuUmpEdgeDeg(AbuUmpEdgeBase, BuyUmpMixin):
    """边裁走势拟合角度特征类，AbuUmpEdgeBase子类，混入BuyUmpMixin，做为买入ump类"""

    class UmpDegFiter(AbuMLPd):
        """
            内部类，只需要实现make_xy，且使用ump_edge_make_xy装饰

            在边裁__init__中通过：
                    self.fiter_cls = self.get_fiter_class()
                    self.fiter = self.fiter_cls(orders_pd=orders_pd, **kwarg)
            构造裁判的filter以及重要的self.fiter.df，即pd.DataFrame对象特征
        """

        @ump_edge_make_xy
        def make_xy(self, **kwarg):
            """
             make_xy通过装饰器ump_edge_make_xy进行包装，调用make_xy前将有交易结果的单子进行筛选：
             order_has_ret = orders_pd[orders_pd['result'] != 0]，且赋予self.order_has_ret
             make_xy只需要使用filter选取需要的特征，即从self.order_has_ret中使用filter选取特征列形成df

            :param kwarg: ump_edge_make_xy装饰器中使用kwarg
                          kwargs['orders_pd'] 做为必须要有的关键字参数：交易训练集数据，pd.DataFrame对象
            :return: self.order_has_ret中使用filter选取特征列形成deg_df
                     ump_edge_make_xy装饰器在make_xy返回deg_df后做转换matrix，形成x，y等工作
            """

            filter_list = ['profit', 'profit_cg']
            cols = AbuFeatureDeg().get_feature_ump_keys(ump_cls=AbuUmpEdgeDeg)
            # filter_list: ['profit', 'profit_cg', 'buy_deg_ang42', 'buy_deg_ang252', 'buy_deg_ang60', 'buy_deg_ang21']
            filter_list.extend(cols)

            # noinspection PyUnresolvedReferences
            deg_df = self.order_has_ret.filter(filter_list)
            """
                eg: deg_df
                               profit  profit_cg  buy_deg_ang42  buy_deg_ang252  \
                2014-09-24  -22618.04    -0.0566          3.378           3.458
                2014-10-24  -29690.28    -0.0742          0.191           2.889
                2014-10-29   18959.19     0.0542         -2.026          16.689
                2014-10-29  148209.36     0.5022         -3.427         -11.956
                2014-10-29   24867.60     0.0952         -2.915          39.469
                2014-10-29   18959.19     0.0542         -2.026          16.689
                2014-11-03    1250.80     0.0045          0.103          39.202
                2014-11-11   59888.21     0.1857          8.341          -9.450
                2014-11-12   -3578.78    -0.0140          3.963           6.595
                2014-11-26  -29085.19    -0.0946         14.052           6.061
                ...               ...        ...            ...             ...
                2016-03-14   16220.57     0.0559          4.002         -10.559
                2016-03-14  -25328.12    -0.1218          0.129          -6.649
                2016-03-30  -29858.44    -0.0863         13.121          -8.461
                2016-04-04    5373.76     0.0244          4.409         -33.097
                2016-04-13  -28044.40    -0.1159          6.603         -31.459
                2016-04-14  -18645.93    -0.0467          4.611          18.428
                2016-04-15  -32484.79    -0.1149          4.238         -13.247
                2016-04-15  -32484.79    -0.1149          4.238         -13.247
                2016-04-29     290.96     0.0007          1.445          16.266
                2016-04-29     290.96     0.0007          1.445          16.266

                            buy_deg_ang60  buy_deg_ang21
                2014-09-24          3.458          1.818
                2014-10-24          2.809         -1.089
                2014-10-29         -0.761          1.980
                2014-10-29         -8.296          6.507
                2014-10-29         -6.043          7.046
                2014-10-29         -0.761          1.980
                2014-11-03         -4.614         10.125
                2014-11-11          0.730         12.397
                2014-11-12         -7.524          6.671
                2014-11-26          7.566         12.494
                ...                   ...            ...
                2016-03-14         -7.992          9.324
                2016-03-14        -10.880          5.201
                2016-03-30          4.498          4.070
                2016-04-04         -6.281          5.618
                2016-04-13          0.191          4.457
                2016-04-14          3.134          0.733
                2016-04-15          4.693          1.162
                2016-04-15          4.693          1.162
                2016-04-29          4.615         -1.115
                2016-04-29          4.615         -1.115
            """
            return deg_df

    def get_predict_col(self):
        """
        边裁走势拟合角度特征keys：['buy_deg_ang42', 'buy_deg_ang252', 'buy_deg_ang60', 'buy_deg_ang21']
        :return: ['buy_deg_ang42', 'buy_deg_ang252', 'buy_deg_ang60', 'buy_deg_ang21']
        """
        return AbuFeatureDeg().get_feature_ump_keys(ump_cls=AbuUmpEdgeDeg)

    def get_fiter_class(self):
        """
        边裁特征走势拟合角度返回的AbuMLPd子类：AbuUmpEdgeDeg.UmpDegFiter
        :return: AbuUmpEdgeDeg.UmpDegFiter
        """
        return AbuUmpEdgeDeg.UmpDegFiter

    @classmethod
    def class_unique_id(cls):
        """
        具体ump类关键字唯一名称，类方法：return 'deg_edge'
        主要针对外部user设置自定义ump使用, 需要user自己保证class_unique_id的唯一性，内部不做检测
        具体使用见ABuUmpManager中extend_ump_block方法
        """
        return 'deg_edge'


class AbuUmpEegeDegExtend(AbuUmpEdgeBase, BuyUmpMixin):
    """边裁使用新的视角来决策交易，AbuUmpEdgeBase子类，混入BuyUmpMixin，做为买入ump类"""

    class UmpExtendEdgeFiter(AbuMLPd):
        @ump_edge_make_xy
        def make_xy(self, **kwarg):
            filter_list = ['profit', 'profit_cg']
            col = AbuFeatureDegExtend().get_feature_ump_keys(ump_cls=AbuUmpEegeDegExtend)
            filter_list.extend(col)
            mul_df = self.order_has_ret.filter(filter_list)
            return mul_df

    def get_predict_col(self):
        # 这里使用get_feature_ump_keys，只需要传递当前类名称即可，其根据是买入ump还是卖出ump返回对应特征列
        col = AbuFeatureDegExtend().get_feature_ump_keys(ump_cls=AbuUmpEegeDegExtend)
        return col

    def get_fiter_class(self):
        return AbuUmpEegeDegExtend.UmpExtendEdgeFiter

    @classmethod
    def class_unique_id(cls):
        return 'extend_edge_deg'



# -*- encoding:utf-8 -*-
"""
    md5, crc32等加密，变换匹配模块
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import hashlib
import os
from binascii import crc32

# noinspection PyUnresolvedReferences
from ..CoreBu.ABuFixes import xrange
from ..CoreBu import ABuEnv

K_BYTE_SIZE = 1024


def _md5_obj():
    """根据python版本返回md5实例"""
    md5_obj = hashlib.md5() if ABuEnv.g_is_py3 else hashlib.new("md5")
    return md5_obj


def md5_from_binary(binary):
    """对字符串进行md5, 返回md5后32位字符串对象"""
    m = _md5_obj()
    m.update(binary.encode('utf-8'))
    return m.hexdigest()


def md5_from_file(fn, block_m=1):
    """
    对文件进行md5, 分块读取文件
    :param fn: 目标文件路径
    :param block_m: 分块读取大小，默认1mb
    :return: md5后32位字符串对象, md5失败等问题返回0
    """
    opened = False
    f_obj = None
    if hasattr(fn, "read"):
        f_obj = fn
    else:
        if os.path.exists(fn) and os.path.isfile(fn):
            f_obj = open(fn, "rb")
            opened = True
    if f_obj:
        block_b = block_m * K_BYTE_SIZE * K_BYTE_SIZE
        try:
            m = _md5_obj()
            while True:
                fb = f_obj.read(block_b)
                if not fb:
                    break
                m.update(fb)
        finally:
            if opened:
                f_obj.close()
        return m.hexdigest()
    else:
        return 0


def crc32_from_file(fn, block_m=1):
    """
    对文件进行crc32, 分块读取文件
    :param fn: 目标文件路径
    :param block_m: 分块读取大小，默认1mb
    :return: crc32后返回的16进制字符串 eg. '0x00000000'
    """
    if os.path.exists(fn) and os.path.isfile(fn):
        block_b = block_m * K_BYTE_SIZE * K_BYTE_SIZE
        crc = 0
        f = open(fn, "rb")
        while True:
            fb = f.read(block_b)
            if not fb:
                break
            crc = crc32(fb, crc)
        f.close()
        res = ''
        for _ in xrange(4):
            t = crc & 0xFF
            crc >>= 8
            res = '%02x%s' % (t, res)
        return "0x" + res
    else:
        return 0



# -*- encoding:utf-8 -*-
"""
    多任务子进程拷贝跟随主进程设置模块
"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import functools

# noinspection PyUnresolvedReferences
from ..CoreBu.ABuFixes import filter
from ..CoreBu.ABuFixes import signature, Parameter

__author__ = '阿布'
__weixin__ = 'abu_quant'


def add_process_env_sig(func):
    """
    初始化装饰器时给被装饰函数添加env关键字参数，在wrapper中将env对象进行子进程copy
    由于要改方法签名，多个装饰器的情况要放在最下面
    :param func:
    :return:
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        # env = kwargs.pop('env', None)
        if 'env' in kwargs:
            """
                实际上linux, mac os上并不需要进行进程间模块内存拷贝，
                子进程fork后携带了父进程的内存信息，win上是需要的，
                暂时不做区分，都进行进程间的内存拷贝，如特别在乎效率的
                情况下基于linux系统，mac os可以不需要拷贝，如下：
                if kwargs['env'] is not None and not ABuEnv.g_is_mac_os:
                    # 只有windows进行内存设置拷贝
                    env.copy_process_env()
            """
            # if kwargs['env'] is not None and not ABuEnv.g_is_mac_os:
            env = kwargs.pop('env', None)
            if env is not None:
                # 将主进程中的env拷贝到子进程中
                env.copy_process_env()
        return func(*args, **kwargs)

    # 获取原始函数参数签名，给并行方法添加env参数
    sig = signature(func)

    if 'env' not in list(sig.parameters.keys()):
        parameters = list(sig.parameters.values())
        # 通过强制关键字参数，给方法加上env
        parameters.append(Parameter('env', Parameter.KEYWORD_ONLY, default=None))
        # wrapper的__signature__进行替换
        wrapper.__signature__ = sig.replace(parameters=parameters)

    return wrapper


class AbuEnvProcess(object):
    """多任务主进程内存设置拷贝执行者类"""

    def __init__(self):
        """迭代注册了的需要拷贝内存设置的模块，通过筛选模块中以g_或者_g_开头的属性，将这些属性拷贝为类属性变量"""
        for module in self.register_module():
            # 迭代注册了的需要拷贝内存设置的模块, 筛选模块中以g_或者_g_开头的, 且不能callable，即不是方法
            sig_env = list(filter(
                lambda _sig: not callable(_sig) and (_sig.startswith('g_') or _sig.startswith('_g_')), dir(module)))

            module_name = module.__name__
            # map(lambda sig: setattr(self, '{}_{}'.format(module_name, sig), module.__dict__[sig]), sig_env)
            for sig in sig_env:
                # 模块中的属性拷贝为类属性变量，key＝module_name_sig
                setattr(self, '{}_{}'.format(module_name, sig), module.__dict__[sig])

    # noinspection PyMethodMayBeStatic
    def register_module(self):
        """
        注册需要拷贝内存的模块，不要全局模块注册，否则很多交叉引用，也不要做为类变量存储否则多进程传递pickle时会出错
        :return:
        """
        from ..BetaBu import ABuAtrPosition, ABuPositionBase
        from ..CoreBu import ABuEnv
        from ..SimilarBu import ABuCorrcoef
        from ..SlippageBu import ABuSlippageBuyMean, ABuSlippageBuyBase, ABuSlippageSellBase
        from ..TradeBu import ABuMLFeature
        from ..UmpBu import ABuUmpManager, ABuUmpMainBase, ABuUmpEdgeBase
        from ..TLineBu import ABuTLSimilar
        from ..MarketBu import ABuMarket
        from ..AlphaBu import ABuPickTimeWorker
        from ..FactorSellBu import ABuFactorCloseAtrNStop, ABuFactorPreAtrNStop
        from ..PickStockBu import ABuPickSimilarNTop
        from ..UtilBu import ABuProgress

        # TODO 将每个模块中全局设置放在一个模块配置代码文件中，这里只将所有模块配置代码文件加载
        return [ABuAtrPosition, ABuPositionBase, ABuEnv, ABuCorrcoef, ABuProgress,
                ABuSlippageBuyMean, ABuSlippageSellBase, ABuSlippageBuyBase, ABuUmpMainBase, ABuUmpEdgeBase,
                ABuMLFeature, ABuUmpManager, ABuTLSimilar, ABuPickTimeWorker,
                ABuFactorCloseAtrNStop, ABuMarket, ABuFactorPreAtrNStop, ABuPickSimilarNTop]

    def copy_process_env(self):
        """为子进程拷贝主进程中的设置执行，在add_process_env_sig装饰器中调用，外部不应主动使用"""
        for module in self.register_module():
            # 迭代注册了的需要拷贝内存设置的模块, 筛选模块中以g_或者_g_开头的, 且不能callable，即不是方法
            sig_env = list(filter(
                lambda sig: not callable(sig) and (sig.startswith('g_') or sig.startswith('_g_')), dir(module)))
            module_name = module.__name__
            for _sig in sig_env:
                # 格式化类变量中对应模块属性的key
                name = '{}_{}'.format(module_name, _sig)
                # 根据应模块属性的key（name）getattr获取属性值
                val = getattr(self, name)
                # 为子模块内存变量进行值拷贝
                module.__dict__[_sig] = val
                # print(name, val)

    def __str__(self):
        """打印对象显示：注册需要拷贝内存的模块中在AbuEnvProcess对象属性的映射key值，以及value值"""

        str_dict = dict()
        for module in self.register_module():
            sig_env = list(filter(
                lambda sig: not callable(sig) and (sig.startswith('g_') or sig.startswith('_g_')), dir(module)))
            module_name = module.__name__
            for _sig in sig_env:
                # format对象属性的映射key值
                name = '{}_{}'.format(module_name, _sig)
                # 根据映射key值getattr出value值
                attr_str = getattr(self, name)
                str_dict[name] = attr_str
        return str(str_dict)

    __repr__ = __str__



# -*- encoding:utf-8 -*-
"""示例ump边裁特征多混模块"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from ..TradeBu.ABuMLFeature import AbuFeatureDeg, AbuFeaturePrice, AbuFeatureWave, AbuFeatureAtr
from ..MLBu.ABuMLPd import AbuMLPd
from .ABuUmpBase import ump_edge_make_xy, BuyUmpMixin
from .ABuUmpEdgeBase import AbuUmpEdgeBase

__author__ = '阿布'
__weixin__ = 'abu_quant'


class AbuUmpEdgeFull(AbuUmpEdgeBase, BuyUmpMixin):
    """边裁多混特征类，AbuUmpEdgeBase子类，混入BuyUmpMixin，做为买入ump类"""

    class UmpFullFiter(AbuMLPd):
        """
            内部类，只需要实现make_xy，且使用ump_edge_make_xy装饰

            在边裁__init__中通过：
                    self.fiter_cls = self.get_fiter_class()
                    self.fiter = self.fiter_cls(orders_pd=orders_pd, **kwarg)
            构造裁判的filter以及重要的self.fiter.df，即pd.DataFrame对象特征
        """

        @ump_edge_make_xy
        def make_xy(self, **kwarg):
            """
             make_xy通过装饰器ump_edge_make_xy进行包装，调用make_xy前将有交易结果的单子进行筛选：
             order_has_ret = orders_pd[orders_pd['result'] != 0]，且赋予self.order_has_ret
             make_xy只需要使用filter选取需要的特征，即从self.order_has_ret中使用filter选取特征列形成df

            :param kwarg: ump_edge_make_xy装饰器中使用kwarg
                          kwargs['orders_pd'] 做为必须要有的关键字参数：交易训练集数据，pd.DataFrame对象
            :return: self.order_has_ret中使用filter选取特征列形成full_df
                     ump_edge_make_xy装饰器在make_xy返回full_df后做转换matrix，形成x，y等工作
            """

            filter_list = ['profit', 'profit_cg']
            cols = AbuFeatureDeg().get_feature_ump_keys(ump_cls=AbuUmpEdgeFull)
            cols.extend(AbuFeaturePrice().get_feature_ump_keys(ump_cls=AbuUmpEdgeFull))
            cols.extend(AbuFeatureWave().get_feature_ump_keys(ump_cls=AbuUmpEdgeFull))
            cols.extend(AbuFeatureAtr().get_feature_ump_keys(ump_cls=AbuUmpEdgeFull))

            filter_list.extend(cols)
            """
                filter_list:
                ['profit', 'profit_cg', 'buy_deg_ang42', 'buy_deg_ang252', 'buy_deg_ang60', 'buy_deg_ang21',
                'buy_price_rank120', 'buy_price_rank90', 'buy_price_rank60', 'buy_price_rank252', 'buy_wave_score1',
                'buy_wave_score2', 'buy_wave_score3', 'buy_atr_std']
            """
            # noinspection PyUnresolvedReferences
            full_df = self.order_has_ret.filter(filter_list)
            """
                eg: full_df
                               profit  profit_cg  buy_deg_ang42  buy_deg_ang252  \
                2014-09-24  -22618.04    -0.0566          3.378           3.458
                2014-10-24  -29690.28    -0.0742          0.191           2.889
                2014-10-29   18959.19     0.0542         -2.026          16.689
                2014-10-29  148209.36     0.5022         -3.427         -11.956
                2014-10-29   24867.60     0.0952         -2.915          39.469
                2014-10-29   18959.19     0.0542         -2.026          16.689
                ...               ...        ...            ...             ...
                2016-04-13  -28044.40    -0.1159          6.603         -31.459
                2016-04-14  -18645.93    -0.0467          4.611          18.428
                2016-04-15  -32484.79    -0.1149          4.238         -13.247
                2016-04-15  -32484.79    -0.1149          4.238         -13.247
                2016-04-29     290.96     0.0007          1.445          16.266
                2016-04-29     290.96     0.0007          1.445          16.266

                            buy_deg_ang60  buy_deg_ang21  buy_price_rank120  buy_price_rank90  \
                2014-09-24          3.458          1.818              1.000             1.000
                2014-10-24          2.809         -1.089              1.000             1.000
                2014-10-29         -0.761          1.980              1.000             1.000
                2014-10-29         -8.296          6.507              0.925             0.900
                2014-10-29         -6.043          7.046              0.962             0.950
                2014-10-29         -0.761          1.980              1.000             1.000
                ...                   ...            ...                ...               ...
                2016-04-13          0.191          4.457              0.567             0.722
                2016-04-14          3.134          0.733              0.875             0.878
                2016-04-15          4.693          1.162              0.775             0.733
                2016-04-15          4.693          1.162              0.775             0.733
                2016-04-29          4.615         -1.115              1.000             1.000
                2016-04-29          4.615         -1.115              1.000             1.000

                            buy_price_rank60  buy_price_rank252  buy_wave_score1  \
                2014-09-24             1.000              1.000            0.287
                2014-10-24             1.000              1.000            0.596
                2014-10-29             1.000              1.000            0.444
                2014-10-29             0.883              0.750           -0.173
                2014-10-29             0.925              0.982            0.031
                2014-10-29             1.000              1.000            0.444
                ...                      ...                ...              ...
                2016-04-13             1.000              0.270            0.271
                2016-04-14             0.967              0.940           -0.030
                2016-04-15             1.000              0.631            0.596
                2016-04-15             1.000              0.631            0.596
                2016-04-29             1.000              1.000            0.743
                2016-04-29             1.000              1.000            0.743

                            buy_wave_score2  buy_wave_score3  buy_atr_std
                2014-09-24            0.234            0.218        0.226
                2014-10-24            0.488            0.449        0.146
                2014-10-29            0.338            0.329        0.538
                2014-10-29           -0.202           -0.203        0.558
                2014-10-29           -0.128           -0.173        0.063
                2014-10-29            0.338            0.329        0.538
                ...                     ...              ...          ...
                2016-04-13            0.509            0.651        0.262
                2016-04-14            0.081            0.170        0.049
                2016-04-15            0.753            0.800        0.135
                2016-04-15            0.753            0.800        0.135
                2016-04-29            0.840            0.918        0.838
                2016-04-29            0.840            0.918        0.838
            """
            return full_df

    def get_predict_col(self):
        """
        边裁多混特征keys：
            ['buy_deg_ang42', 'buy_deg_ang252', 'buy_deg_ang60', 'buy_deg_ang21', 'buy_price_rank120',
            'buy_price_rank90', 'buy_price_rank60', 'buy_price_rank252',
            'buy_wave_score1', 'buy_wave_score2', 'buy_wave_score3', 'buy_atr_std',
            'buy_diff_down_days', 'buy_jump_up_power', 'buy_diff_up_days', 'buy_jump_down_power']

        :return: ['buy_deg_ang42', 'buy_deg_ang252', 'buy_deg_ang60', 'buy_deg_ang21', 'buy_price_rank120',
                'buy_price_rank90', 'buy_price_rank60', 'buy_price_rank252',
                'buy_wave_score1', 'buy_wave_score2', 'buy_wave_score3', 'buy_atr_std',
                'buy_diff_down_days', 'buy_jump_up_power', 'buy_diff_up_days', 'buy_jump_down_power']
        """

        cols = AbuFeatureDeg().get_feature_ump_keys(ump_cls=AbuUmpEdgeFull)
        cols.extend(AbuFeaturePrice().get_feature_ump_keys(ump_cls=AbuUmpEdgeFull))
        cols.extend(AbuFeatureWave().get_feature_ump_keys(ump_cls=AbuUmpEdgeFull))
        cols.extend(AbuFeatureAtr().get_feature_ump_keys(ump_cls=AbuUmpEdgeFull))
        return cols

    def get_fiter_class(self):
        """
        边裁多混特征返回的AbuMLPd子类：AbuUmpEdgeFull.UmpFullFiter
        :return: AbuUmpEdgeFull.UmpFullFiter
        """
        return AbuUmpEdgeFull.UmpFullFiter

    @classmethod
    def class_unique_id(cls):
        """
        具体ump类关键字唯一名称，类方法：return 'full_edge'
        主要针对外部user设置自定义ump使用, 需要user自己保证class_unique_id的唯一性，内部不做检测
        具体使用见ABuUmpManager中extend_ump_block方法
        """
        return 'full_edge'



# -*- encoding:utf-8 -*-
"""量化技术分析工具图形可视化"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import ipywidgets as widgets

from ..WidgetBu.ABuWGBase import WidgetBase
from ..WidgetBu.ABuWGToolBase import WidgetToolSet
from ..WidgetBu.ABuWGTLTool import WidgetTLTool
from ..WidgetBu.ABuWGDATool import WidgetDATool
from ..WidgetBu.ABuWGSMTool import WidgetSMTool

__author__ = '阿布'
__weixin__ = 'abu_quant'


class WidgetQuantTool(WidgetBase):
    """量化分析工具主界面"""

    def __init__(self):
        self.ts = WidgetToolSet()
        self.da = WidgetDATool(self.ts)
        self.tl = WidgetTLTool(self.ts)
        self.sm = WidgetSMTool(self.ts)
        sub_widget_tab = widgets.Tab()
        sub_widget_tab.children = [self.tl.widget, self.sm.widget, self.da.widget, self.ts.widget]
        for ind, name in enumerate([u'技术分析', u'相关分析', u'振幅分析', u'分析设置']):
            sub_widget_tab.set_title(ind, name)
        self.widget = widgets.VBox([sub_widget_tab])
